import logging
from ibmsecurity.isam.aac.policy_information_points.all import get, search, _create_json
from ibmsecurity.utilities.tools import json_sort

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/pips"
requires_modules = ["mga"]
requires_version = None


def add(isamAppliance, name, properties, attributes, description=None, type="Database", check_mode=False,
        force=False):
    """
    Create a Database policy information point
    """

    ret_obj = search(isamAppliance, name, check_mode=False, force=False)
    id = ret_obj['data']

    if id != {}:
        logger.info(f"PIP '{name}' already exists.  Skipping add.")

    if force is True or id == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_post(
                "Create a Database policy information point",
                f"{uri}",
                _create_json(name=name, description=description, type=type,
                             attributes=attributes, properties=properties),
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def update(isamAppliance, name, properties, attributes, description=None, type="Database", new_name=None,
           check_mode=False, force=False):
    """
    Update a specific Database policy information point
    """
    ret_obj = search(isamAppliance, name=name)
    id = ret_obj['data']

    update_required = False

    if id != {}:
        ret_obj = get(isamAppliance, name=name)

        if new_name != None:
            json_data = _create_json(name=new_name, properties=properties, attributes=attributes,
                                     description=description, type=type)
        else:
            json_data = _create_json(name=name, properties=properties, attributes=attributes, description=description,
                                     type=type)

        sorted_json_data = json_sort(json_data)

        logger.debug(f"Sorted input: {sorted_json_data}")

        del ret_obj['data']['id']
        del ret_obj['data']['predefined']
        sorted_ret_obj = json_sort(ret_obj['data'])

        logger.debug(f"Sorted existing data: {sorted_ret_obj}")

        if sorted_json_data != sorted_ret_obj:
            update_required = True
    else:
        logger.info(f"PIP '{name}' does not exists.  Skipping update.")
        warnings = [f"PIP '{name}' does not exists.  Skipping update."]
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)

        else:

            return isamAppliance.invoke_put(
                "Update a specific Database policy information point",
                f"{uri}/{id}",
                json_data,
                requires_modules=requires_modules, requires_version=requires_version
            )

    if update_required is False:
        logger.info(f"Input is the same as current PIP '{name}'.  Skipping update.")

    return isamAppliance.create_return_object()


def set(isamAppliance, name, properties, attributes, description=None, type="Database", new_name=None, check_mode=False,
        force=False):
    """
    Creating or Modifying a Database PIP
    """
    ret_obj = search(isamAppliance, name=name)
    id = ret_obj['data']

    if id == {}:
        # If no id was found, Force the add
        return add(isamAppliance, name=name, properties=properties, attributes=attributes, description=description,
                   type=type, check_mode=check_mode, force=force)
    else:
        # Update PIP
        return update(isamAppliance, name=name, properties=properties, attributes=attributes, description=description,
                      type=type, new_name=new_name, check_mode=check_mode, force=force)



import ibmsecurity.utilities.tools
import logging

module_uri = "/isam/felb/configuration/ssl"
requires_module = None
requires_version = None
requires_model = "Appliance"

logger = logging.getLogger(__name__)


def enable(isamAppliance, keyfile, check_mode=False, force=False):
    """
    Creating SSL configuration
    """

    check_enable, warnings = _check_enable(isamAppliance, keyfile)
    if force is True or check_enable is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post("Creating SSL configuration", module_uri,
                                             {
                                                 "keyfile": keyfile
                                             },
                                             requires_version=requires_version,
                                             requires_modules=requires_module,
                                             requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def disable(isamAppliance, check_mode=False, force=False):
    """
    Deletes ssl configuration
    """

    check_disable, warnings = _check_disable(isamAppliance)
    if force is True or check_disable is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete("Disabling SSL", module_uri,
                                               requires_version=requires_version,
                                               requires_modules=requires_module,
                                               requires_model=requires_model)
    else:
        return isamAppliance.create_return_object(warnings=warnings)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieves ssl configuration
    """
    return isamAppliance.invoke_get("Retrieving SSL configuration", module_uri,
                                    requires_version=requires_version,
                                    requires_modules=requires_module,
                                    requires_model=requires_model)


def update(isamAppliance, keyfile, check_mode=False, force=False):
    """
    Updating SSL configuration
    """
    # Call to check function to see if configuration already exist
    update_required, warnings = _check_enable(isamAppliance, keyfile)

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put("Updating SSL configuration", module_uri,
                                            {
                                                'keyfile': keyfile

                                            },
                                            requires_modules=requires_module,
                                            requires_version=requires_version,
                                            requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_enable(isamAppliance, keyfile=None):
    """
    checks add function for idempotency
    """
    change_required = False
    check_obj = get(isamAppliance)
    warnings = check_obj['warnings']
    # checks to see if ssl configuration exists

    if 'enabled' in check_obj['data']:
        if check_obj['data']['enabled'] is False:
            return True, warnings
        elif check_obj['data']['keyfile'] != keyfile:
            return True, warnings
        else:
            return False, warnings
    else:
        return False, warnings


def _check_disable(isamAppliance):
    """
    Checks delete function for idempotency
    """

    check_obj = get(isamAppliance)
    warnings = check_obj['warnings']

    if 'enabled' in check_obj['data']:
        if check_obj['data']['enabled'] == True:
            return True, warnings
        else:
            return False, warnings
    else:
        return False, warnings


def compare(isamAppliance1, isamAppliance2):
    """
    Compare cluster configuration between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)
artifact_type = "Obligation type"

# URI for this module
uri = "/iam/access/v8/obligation/types"


def search(isamAppliance, name):
    """
    Search for Obligation Type by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj["data"]:
        if obj["name"] == name:
            logger.info(f"Found {artifact_type} \"{name}\" with id: {obj['id']}")
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def get_all(isamAppliance, start=None, count=None, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of Obligation types
    """
    return isamAppliance.invoke_get("Retrieve a list of {0}", "{1}/{2}".format(artifact_type, uri,
                                                                               tools.create_query_string(start=start,
                                                                                                         count=count,
                                                                                                         filter=filter,
                                                                                                         sortBy=sortBy)))


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific Obligation
    """

    ret_obj = search(isamAppliance, name)

    if ret_obj['data'] == {}:
        logger.info(f'{artifact_type} "{name}" had no match, skipping retrieval.')
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieve a specific obligation",
                                        f"{uri}/{ret_obj['data']}")



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/isam/user_count/user_count.json"
requires_modules = ["wga"]
requires_version = '11.0.2.0'


def get(isamAppliance, use_cache_result=False, search_timeout_sec=30, check_mode=False, force=False):
    """
    Get the user count
    """
    if use_cache_result:
        use_cache_result = 'true'
    else:
        use_cache_result = 'false'

    return isamAppliance.invoke_get("Retrieve the user count",
                                    f"{uri}?use_cache_result={use_cache_result}&search_timeout_sec={search_timeout_sec}",
                                        requires_modules=requires_modules,
                                        requires_version=requires_version
                                    )



import logging
import ibmsecurity.utilities.tools
import json

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Get management authentication
    """
    return isamAppliance.invoke_get("Get management authentication",
                                    "/isam/management_authentication/")


def set(isamAppliance,
        check_mode=False,
        force=False,
        ignore_password_for_idempotency=False,
        type='remote',
        **kwargs,
       ):
    """
    Set management authentication to remote

    ldap_host=None,
    ldap_port=None,
    base_dn=None,
    admin_group_dn=None,
    enable_ssl=False,
    key_database=None,
    cert_label=None,
    user_attribute='uid',
    group_member_attribute='member',
    anon_bind=True,
    bind_dn=None,
    bind_password=None,
    ldap_debug=None,
    enable_usermapping=False,
    usermapping_script=None,
    enable_ssh_pubkey_auth
    ssh_pubkey_auth_attribute=None
    oidc_client_id	String	No	The OIDC Client Identifier. This field is required if type == "federation".
    oidc_client_secret	String	No	The OIDC Client Secret. This field is required if type == "federation".
    oidc_discovery_endpoint	String	No	The OIDC Discovery (well-known) endpoint. This field is required if type == "federation".
    oidc_enable_pkce	Boolean	No	Specifies whether the Public key Code Exchange extension is enforced. This field is required if type == "federation".
    oidc_enable_admin_group	Boolean	No	Specifies whether a user must be a member of a particular group to be considered an administrator user. This field is required if type == "federation".
    oidc_group_claim	String	Yes	The OIDC token claim to use as group membership. This claim can either be a String, or a list of Strings. The default value is "groups".
    oidc_admin_group	String	Yes	The name of the group which a user must be a member of to be considered an administrator user. The default value is "adminGroup".
    oidc_user_claim	String	Yes	Specifies the OIDC token claim to use as the username. The default value is "sub".
    oidc_keystore	String	Yes	The SSL Truststore to verify connections the the OIDC OP. The default value if "lmi_trust_store".
    enable_tokenmapping	Boolean	No	Specifies whether custom claim to identity mapping is performed using a JavaScript code fragment. This field is required if type == "federation".
    tokenmapping_script	String	Yes	The custom JavaScript code fragment to map an identity token to a username/group membership.

    """
    warnings = []
    update_required = False

    json_data = {
        'type': type
    }
    for k, v in kwargs.items():
        if k  == 'ldap_debug' and v == '':
            if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "9.0.4.0") < 0:
                warnings.append(
                    f"Appliance at version: {isamAppliance.facts['version']}, ldap_debug: {ldap_debug} is not supported. Needs 9.0.4.0 or higher. Ignoring ldap_debug for this call.")
                continue
        if k == 'enable_usermapping':
            # only valid if type == remote
            if type == 'remote':
                if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.2.0") < 0:
                    warnings.append(
                        f"Appliance at version: {isamAppliance.facts['version']}, {k}: {v} is not supported. Needs 10.0.2.0 or higher. Ignoring {k} for this call.")
                    continue
            else:
                warnings.append(f"{k} is only supported for type remote. Ignoring it for now.")
                continue
        if k in ('ssh_pubkey_auth_attribute','enable_ssh_pubkey_auth'):
            if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.6") < 0:
                warnings.append(f"Appliance at version: {isamAppliance.facts['version']}, {k}: {v} is not supported. Needs 10.0.8.0 or higher. Ignoring {k} for this call.")
                continue
        if k.startswith('oidc_') or k in ('enable_tokenmapping', 'tokenmapping_script'):
            if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.8") < 0:
                warnings.append(f"Appliance at version: {isamAppliance.facts['version']}, {k}: {v} is not supported. Needs 10.0.8.0 or higher. Ignoring {k} for this call.")
                continue
        json_data[k] = v
    # Set defaults for type remote and cleanup
    if type == 'remote':
        if json_data.get('user_attribute', None) is None:
            json_data['user_attribute'] = 'uid'
        if json_data.get('group_member_attribute', None) is None:
            json_data['group_member_attribute'] = 'member'
        if json_data.get('anon_bind', None) is None:
            json_data['anon_bind'] = True
        if not json_data.get('enable_usermapping', False):
           json_data["usermapping_script"] = ''

    # Defaults for type federation (new in 10.0.8) and cleanup
    if type == 'federation':
        if json_data.get('oidc_group_claim', None) is None:
            json_data['oidc_group_claim'] = 'groups'
        if json_data.get('oidc_admin_group', None) is None:
            json_data['oidc_admin_group'] = 'adminGroup'
        if json_data.get('oidc_user_claim', None) is None:
            json_data['oidc_user_claim'] = "sub"
        if json_data.get('oidc_keystore', None) is None:
            json_data['oidc_keystore'] = "lmi_trust_store"
        if not json_data.get('enable_tokenmapping', False):
            json_data["tokenmapping_script"] = ''

    if not force:
        if not ignore_password_for_idempotency and json_data.get('bind_password', None) is not None:
            warnings.append("Unable to read existing bind password to check idempotency.  You can disable this behaviour by passing `ignore_password_for_idempotency=True`")
            update_required = True
        else:
            ret_obj = get(isamAppliance)
            json_data_compare = json_data.copy()
            if json_data.get("bind_dn", None) is None:
                ret_obj["data"].pop("bind_dn", None)
            ret_obj["data"].pop("bind_password", None)
            json_data_compare.pop("bind_password", None)
            sorted_json_data = json.dumps(json_data_compare, skipkeys=True, sort_keys=True)
            logger.debug(f"\n\nSorted input:         {sorted_json_data}")
            sorted_ret_obj = json.dumps(ret_obj['data'], skipkeys=True, sort_keys=True)
            logger.debug(f"\n\nSorted existing data: {sorted_ret_obj}")
            if sorted_ret_obj != sorted_json_data:
                logger.info("Changes detected, update needed.")
                update_required = True

    if force or update_required:
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                f"Set management authentication to {type}",
                "/isam/management_authentication/",
                json_data, warnings=warnings)

    return isamAppliance.create_return_object()


def disable(isamAppliance, check_mode=False, force=False):
    """
    Disable remote management authentication
    """
    return set(isamAppliance=isamAppliance, ldap_host=None, ldap_port=None, base_dn=None, admin_group_dn=None,
               type='local', enable_ssl=None, key_database=None, cert_label=None, user_attribute=None,
               group_member_attribute=None, anon_bind=None, bind_dn=None, bind_password=None, check_mode=check_mode,
               force=force)


def test(isamAppliance, userid, password, check_mode=False, force=False):
    """
    Testing the management authentication
    """
    ret_obj = isamAppliance.invoke_post("Testing the management authentication",
                                        "/isam/management_authentication/",
                                           {
                                            'user': userid,
                                            'password': password
                                           }
                                       )
    ret_obj['changed'] = False

    return ret_obj


def compare(isamAppliance1, isamAppliance2):
    """
    Compare management authentication settings
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import os.path
from ibmsecurity.utilities import tools
from ibmsecurity.utilities.tools import files_same, get_random_temp_dir, json_compare
import shutil

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/bundles"


def get_all(isamAppliance, filter=None, start=None, count=None, check_mode=False, force=False):
    """
    Retrieve a list of bundles
    """
    return isamAppliance.invoke_get("Retrieve a list of bundles",
                                    "{0}{1}".format(uri,
                                                    tools.create_query_string(filter=filter, start=start, count=count)))


def get(isamAppliance, filename, check_mode=False, force=False):
    """
    Retrieve a specific bundle
    """
    ret_obj = search(isamAppliance, filename=filename, check_mode=check_mode, force=force)
    bundle_id = ret_obj['data']

    if bundle_id == {}:
        logger.info(f"Bundle {filename} had no match, skipping retrieval.")
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieve a specific bundle",
                                        f"{uri}/{bundle_id}")


def search(isamAppliance, filename, force=False, check_mode=False):
    """
    Search bundle id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['filename'] == filename:
            logger.info(f"Found Bundle {filename} id: {obj['id']}")
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def create(isamAppliance, filename, check_mode=False, force=False):
    """
    Create a bundle
    """
    (d, f) = os.path.split(filename)
    if force is True or (search(isamAppliance, filename=f))['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Create a bundle", uri,
                {
                    "filename": f
                })

    return isamAppliance.create_return_object()


def verify(isamAppliance, filename, check_mode=False, force=False):
    """
    Verify the contents of a bundle file

    Note: No changes are being made, so changed flag is being forced to False.
    """
    (d, f) = os.path.split(filename)
    ret_obj = isamAppliance.invoke_post_files(
        "Verify the contents of a bundle file",
        f"{uri}/verify",
        [
            {
                'file_formfield': 'file',
                'filename': filename,
                'mimetype': 'application/jar'
            }
        ],
        {
            "import_file": f
        })

    ret_obj['changed'] = False

    return ret_obj


def export_bundle(isamAppliance, filename, extract_filename, check_mode=False, force=False):
    """
    Export a specific bundle
        -filename is the name of the jar on the Appliance (e.g. mybundle.jar)
        -extract_filename is file system location to export the file (e.g. /tmp/mybundle.jar)
    """
    ret_obj = search(isamAppliance, filename)
    if force is True or ret_obj['data'] != {}:  # ret_obj['data'] is the numeric id assigned by the system
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Export a specific bundle",
                f"{uri}/{ret_obj['data']}/file",
                extract_filename)

    return isamAppliance.create_return_object()


def import_bundle(isamAppliance, filename, check_mode=False, force=False):
    """
    Import the bundle file for a bundle

    ret_obj used in the code below looks like this:
        { 'changed': False,
          'data': [ { u'extensions': [ { u'id': u'NewAuthMech',
                                         u'name': u'NewAuthMech',
                                         u'type': u'Authentication Mechanism'}],
                      u'filename': u'newauth.jar',
                      u'id': u'2'}],
          'rc': 0,
          'warnings': []}

    """
    f = os.path.basename(filename)
    warnings, bundle_id = _check_import(isamAppliance, filename)
    if force is True or bundle_id:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post_files(
                "Import the bundle file for a bundle",
                f"{uri}/{bundle_id}/file",
                [
                    {
                        'file_formfield': 'file',
                        'filename': filename,
                        'mimetype': 'application/jar'
                    }
                ],
                {
                    "import_file": f
                }, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_import(isamAppliance, filename):
    """
    Checks if file on the Appliance exists and if so, whether it is different from filename
    """
    warnings = []
    (d, f) = os.path.split(filename)  # this means the name of the bundle has to match name of file to import
    ret_obj = get(isamAppliance, f)
    tmpdir = get_random_temp_dir()
    if ret_obj['data'] != {} and ret_obj['data']['extensions'] != []:
        tmp_original_file = os.path.join(tmpdir, ret_obj['data']['filename'])
        export_bundle(isamAppliance, ret_obj['data']['filename'], tmp_original_file, check_mode=False, force=True)
        logger.debug("file already exists on appliance")
        if files_same(tmp_original_file, filename):
            logger.debug("files are the same, so we don't want to do anything")
            shutil.rmtree(tmpdir)
            return warnings, False
        else:
            logger.debug("files are different, so we replace existing file")
            bundle_id = ret_obj['data']['id']
            shutil.rmtree(tmpdir)
            return warnings, bundle_id
    elif ret_obj['data'] != {} and ret_obj['data']['extensions'] == []:
        bundle_id = ret_obj['data']['id']
        return warnings, bundle_id
    else:
        warnings.append("Bundle does not exist on appliance, create a bundle first and then import.")
        return warnings, False


def _check(isamAppliance, id):
    """
    Check if id already exists
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['id'] == id:
            return True

    return False


def delete(isamAppliance, filename, check_mode=False, force=False):
    """
    Delete a bundle
    """
    (d, f) = os.path.split(filename)
    ret_obj = search(isamAppliance, f)
    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a bundle",
                f"{uri}/{ret_obj['data']}")

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare list of extension bundles between 2 appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/policyattachments"


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of configured resources
    """
    return isamAppliance.invoke_get(
        "Retrieve a list of configured resources",
        f"{uri}{tools.create_query_string(filter=filter, sortBy=sortBy)}",
    )


def get(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Retrieve a specific configured resource
    """
    ret_obj = search(
        isamAppliance,
        server=server,
        resourceUri=resourceUri,
        check_mode=check_mode,
        force=force,
    )
    resource_id = ret_obj.get("data", {})

    if resource_id == {}:
        logger.info(
            f"Resource {server}{resourceUri} had no match, skipping retrieval."
        )
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get(
            "Retrieve a specific configured resource",
            f"{uri}/{resource_id}",
        )


def get_attachments(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Retrieve a list of attachments for a resource
    """
    ret_obj = search(
        isamAppliance,
        server=server,
        resourceUri=resourceUri,
        check_mode=check_mode,
        force=force,
    )
    resource_id = ret_obj["data"]

    if resource_id == {}:
        logger.info(
            f"Resource {server}{resourceUri} had no match, skipping retrieval."
        )
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get(
            "Retrieve a list of attachments for a resource",
            f"{uri}/{resource_id}/policies",
        )


def search(isamAppliance, server, resourceUri, force=False, check_mode=False):
    """
    Search server/resource uri id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj["data"]:
        if obj["resourceUri"] == resourceUri and obj["server"] == server:
            logger.info(
                f"Found server/resourceUri {server}{resourceUri} id: {obj['id']}"
            )
            return_obj["data"] = obj["id"]
            return_obj["rc"] = 0
            return return_obj

    return return_obj


def authenticate(
    isamAppliance, username, password, domain="Default", check_mode=False, force=False
):
    """
    Authenticate with Security Access Manager
    """
    ret_obj = isamAppliance.invoke_post(
        "Authenticate with Security Access Manager",
        f"{uri}/pdadmin",
        {
            "command": "setCredential",
            "username": username,
            "password": password,
            "domain": domain,
        },
    )

    ret_obj["changed"] = (
        False  # Authentication call changes nothing on the appliance (HTTP POST defaults to change)
    )

    return ret_obj


def get_resources(isamAppliance, object="/WebSEAL", check_mode=False, force=False):
    """
    Retrieve a list of resources in a protected object space
    """
    ret_obj = isamAppliance.invoke_post(
        "Retrieve a list of resources in a protected object space",
        f"{uri}/pdadmin",
        {"command": "object list", "object": object},
    )

    ret_obj["changed"] = (
        False  # Resource get call changes nothing on the appliance (HTTP POST defaults to change)
    )

    return ret_obj


def config(
    isamAppliance,
    server,
    resourceUri,
    policies=[],
    policyType=None,
    policyCombiningAlgorithm=None,
    cache=None,
    check_mode=False,
    force=False,
):
    """
    Configure a resource
    Note: Please input policies with policy names (it will be converted to id's), like so:
     [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
      {'name': '<definition name>', 'type': 'definition'}]
    """
    warnings = []
    ret_obj = search(isamAppliance, server, resourceUri)

    json_data = {"server": server, "resourceUri": resourceUri}
    if policyType is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.6.0") < 0:
            warnings.append(
                f"Appliance at version: {isamAppliance.facts['version']}, policyType: {policyType} is not supported. Needs 9.0.6.0 or higher. Ignoring policyType for this call."
            )
        else:
            json_data["type"] = policyType

    json_data["policies"] = _convert_policy_name_to_id(isamAppliance, policies)
    if policyCombiningAlgorithm is not None:
        if policyType is not None and policyType == "application":
            warnings.append(
                "Do not set policy combining algorithm for application type"
            )
        else:
            json_data["policyCombiningAlgorithm"] = policyCombiningAlgorithm
    if cache is not None:
        if tools.version_compare(isamAppliance.facts["version"], "9.0.3.0") < 0:
            warnings.append(
                f"Appliance at version: {isamAppliance.facts['version']}, cache: {cache} is not supported. Needs 9.0.3.0 or higher. Ignoring cache for this call."
            )
        else:
            json_data["cache"] = int(cache)

    if force or ret_obj.get("data", {}) == {}:
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                "Configure a resource", uri, json_data, warnings=warnings
            )
    else:
        # Idempotency check
        logger.debug(f"Idempotency check for {server}{resourceUri}")
        curConfig = get(isamAppliance, server, resourceUri)
        curConfig = curConfig.get("data", {})
        # Need to get rid of lastmodified and name in each policy
        _policies = list(map(lambda item: item.pop('lastmodified', None), curConfig.get("policies", [])))
        _policies = list(map(lambda item: item.pop('name', None), curConfig.get("policies", [])))
        if tools.json_equals(curConfig, json_data):
            # No updates needed
            return isamAppliance.create_return_object()
        else:
            if check_mode:
                return isamAppliance.create_return_object(changed=True, warnings=warnings)
            else:
                # delete and add again
                delete(isamAppliance, server, resourceUri)
                return isamAppliance.invoke_post(
                    "Reconfigure a resource", uri, json_data, warnings=warnings
                )

def update(
    isamAppliance,
    server,
    resourceUri,
    policyCombiningAlgorithm,
    cache=None,
    check_mode=False,
    force=False,
):
    """
    Update the policy attachment combining algorithm
    """
    warnings = []
    ret_obj = get(isamAppliance, server, resourceUri)

    if force or (
        ret_obj["data"] != {} and (
            ret_obj["data"]["policyCombiningAlgorithm"] != policyCombiningAlgorithm or ret_obj["data"]["cache"] != cache
        )
    ):
        json_data = {"policyCombiningAlgorithm": policyCombiningAlgorithm}
        if cache is not None:
            if tools.version_compare(isamAppliance.facts["version"], "9.0.3.0") < 0:
                warnings.append(
                    f"Appliance at version: {isamAppliance.facts['version']}, cache: {cache} is not supported. Needs 9.0.3.0 or higher. Ignoring cache for this call."
                )
            else:
                json_data["cache"] = int(cache)
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Update the policy attachment combining algorithm and cache",
                f"{uri}/{ret_obj['data']['id']}/properties",
                json_data,
                warnings=warnings,
            )

    return isamAppliance.create_return_object()


def update_attachments(
    isamAppliance,
    server,
    resourceUri,
    attachments,
    action,
    check_mode=False,
    force=False,
):
    """
    Update the attachments for a resource

    Provide attachemnts like so:
     [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
      {'name': '<definition name>', 'type': 'definition'}]
    """
    ret_obj = get(isamAppliance, server, resourceUri)
    cur_policies = ret_obj["data"]["policies"]

    if force is True or _check(cur_policies, attachments, action) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            new_policies = _convert_policy_name_to_id(isamAppliance, attachments)
            return isamAppliance.invoke_put(
                "Update the attachments for a resource",
                f"{uri}/{ret_obj['data']['id']}/policies{tools.create_query_string(action=action)}",
                new_policies,
            )
    return isamAppliance.create_return_object()


def _check(policies, attachments, action):
    # Check and see if there is even one match
    full_match = True
    partial_match = False
    for new_pol in attachments:
        match = False
        for cur_pol in policies:
            if (
                new_pol["name"] == cur_pol["name"] and new_pol["type"] == cur_pol["type"]
            ):
                logger.info(
                    f"At least one policy already exists {new_pol['type']}/{new_pol['name']}"
                )
                match = True
                break
        if match is False:
            full_match = False
        elif partial_match is False:
            partial_match = True

    # Add will be rejected if there is even one match
    if action == "add":
        logger.info(
            f"Check if there is at least one match returned: {partial_match}"
        )
        return not partial_match
    # Delete will be rejected if there is a partial match (has to be full match)
    elif action == "remove":
        logger.info(
            f"Check if there is a full match of provided policies: {full_match}"
        )
        return full_match
    # Set requires that there be a full match and the number of elements match
    elif action == "set":
        if len(policies) == len(attachments) and full_match is True:
            return False
        else:
            return True
    # Force delete will be rejected if there is no match
    elif action == "force_remove":
        logger.info(
            f"Check if there is at least one match returned: {partial_match}"
        )
        return partial_match
    else:
        from ibmsecurity.appliance.ibmappliance import IBMError

        raise IBMError("999", f"Unknown action provided: {action}")


def publish(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Publish the policy attachments for a resource
    """
    ret_obj = get(isamAppliance, server, resourceUri)

    deploy_required = False
    # skip application type for deployment.
    if (ret_obj.get("data", {}).get("type", "reverse_proxy") == "reverse_proxy") and (ret_obj.get("data", {}).get("deployrequired", False) or not ret_obj.get("data", {}).get("deployed", False)):
        deploy_required = True

    logger.debug(f"\n\nDeploy required: {deploy_required}\n\n")

    resource_id = ret_obj.get('data', {}).get('id', None)
    logger.debug(f"\n\nID: {resource_id}\n\n")
    # if (force or deploy_required) and ret_obj.get("data", {}).get("policies", []) != []:
    if (force or deploy_required) and resource_id is not None:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Publish the policy attachments for a resource",
                f"{uri}/deployment/{resource_id}",
                {},
            )

    return isamAppliance.create_return_object()


def publish_list(isamAppliance, attachments, check_mode=False, force=False):
    """
    Publish a list of policy attachments

    Note: provide attachments like so:
    [{'server': '<server1>', 'resourceUri': '<resourceuri1>'}, {'server': '<server2>', 'resourceUri': '<resourceuri2>'}]
    """
    id_list = []
    for attach in attachments:
        ret_obj = get(isamAppliance, attach["server"], attach["resourceUri"])
        if force or ret_obj["data"]["deployrequired"]:
            id_list.append(ret_obj["data"]["id"])
    logger.debug(f"Attachments: {id_list}")

    if len(id_list) > 0:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Publish a list of policy attachments",
                f"{uri}/deployment",
                {"policyAttachmentIds": ",".join(id_list)},
            )

    return isamAppliance.create_return_object()


def _convert_policy_name_to_id(isamAppliance, policies):
    """
    Converts this:
    [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
     {'name': '<definition name>}, 'type': 'definition'}]
    to:
    [{'id': '<policy id>', 'type': 'policy'}, {'id': '<policyset id>', 'type': 'policyset'},
     {'id': '<definition id>, 'type': 'definition'}]
    """
    pol_ids = []
    import ibmsecurity.isam.aac.access_control.policies
    import ibmsecurity.isam.aac.access_control.policy_sets
    import ibmsecurity.isam.aac.api_protection.definitions

    for pol in policies:
        name = pol["name"]
        type = pol["type"]
        if type == "policy":
            ret_obj = ibmsecurity.isam.aac.access_control.policies.search(
                isamAppliance, name
            )
            pol_id = ret_obj["data"]
            if pol_id != {}:
                logger.debug(f"Converting policy {name} to ID: {pol_id}")
            else:
                logger.warning(f"Unable to find policy {name}, skipping.")
        elif type == "policyset":
            ret_obj = ibmsecurity.isam.aac.access_control.policy_sets.search(
                isamAppliance, name
            )
            pol_id = ret_obj["data"]
            if pol_id != {}:
                logger.debug(
                    f"Converting policy set {name} to ID: {pol_id}"
                )
            else:
                logger.warning(f"Unable to find policy set {name}, skipping.")
        elif type == "definition":
            ret_obj = ibmsecurity.isam.aac.api_protection.definitions.search(
                isamAppliance, name
            )
            pol_id = ret_obj["data"]
            if pol_id != {}:
                logger.debug(
                    f"Converting api definition {name} to ID: {pol_id}"
                )
            else:
                logger.warning(
                    f"Unable to find api definition {name}, skipping."
                )
        else:
            from ibmsecurity.appliance.ibmappliance import IBMError

            raise IBMError(
                "999", f"Policy specified with unknown type: {type}/{name}"
            )

        pol_ids.append({"id": pol_id, "type": type})

    return pol_ids


def _convert_policy_id_to_name(isamAppliance, policies):
    """
    Converts this:
    [{'id': '<policy id>', 'type': 'policy'}, {'id': '<policyset id>', 'type': 'policyset'},
     {'id': '<definition id>, 'type': 'definition'}]
    to:
    [{'name': '<policy name>', 'type': 'policy'}, {'name': '<policyset name>', 'type': 'policyset'},
     {'name': '<definition name>}, 'type': 'definition'}]
    """
    pol_ids = []
    import ibmsecurity.isam.aac.access_control.policies
    import ibmsecurity.isam.aac.access_control.policy_sets
    import ibmsecurity.isam.aac.api_protection.definitions

    for pol in policies:
        pol_id = pol["id"]
        type = pol["type"]
        if type == "policy":
            ret_obj = ibmsecurity.isam.aac.access_control.policies._get(
                isamAppliance, pol_id
            )
            pol_name = ret_obj["data"]["name"]
            if pol_name != {}:
                logger.debug(
                    f"Converting policy {pol_id} to Name: {pol_name}"
                )
            else:
                logger.warning(f"Unable to find policy {pol_id}, skipping.")
        elif type == "policyset":
            ret_obj = ibmsecurity.isam.aac.access_control.policy_sets._get(
                isamAppliance, pol_id
            )
            pol_name = ret_obj["data"]["name"]
            if pol_name != {}:
                logger.debug(
                    f"Converting policy set {pol_id} to Name: {pol_name}"
                )
            else:
                logger.warning(
                    f"Unable to find policy set {pol_id}, skipping."
                )
        elif type == "definition":
            ret_obj = ibmsecurity.isam.aac.api_protection.definitions._get(
                isamAppliance, pol_id
            )
            pol_name = ret_obj["data"]["name"]
            if pol_name != {}:
                logger.debug(
                    f"Converting api definition {pol_id} to Name: {pol_name}"
                )
            else:
                logger.warning(
                    f"Unable to find api definition {pol_id}, skipping."
                )
        else:
            from ibmsecurity.appliance.ibmappliance import IBMError

            raise IBMError(
                "999",
                f"Policy specified with unknown type: {type}/{pol_id}",
            )

        pol_ids.append({"name": pol_name, "type": type})

    return pol_ids


def delete(isamAppliance, server, resourceUri, check_mode=False, force=False):
    """
    Delete a configured resource
    """
    ret_obj = search(
        isamAppliance, server, resourceUri, check_mode=check_mode, force=force
    )
    resourceID = ret_obj["data"]

    if resourceID == {}:
        logger.info(
            f"ResourceURI {server}/{resourceUri} not found, skipping delete."
        )
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a configured resource", f"{uri}/{resourceID}"
            )

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Policy Sets between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1["data"]:
        del obj["id"]
        del obj["userlastmodified"]
        del obj["lastmodified"]
        del obj["datecreated"]
        del obj["lastdeployed"]
        for pol in obj["policies"]:
            del pol["id"]
            del pol["lastmodified"]
    for obj in ret_obj2["data"]:
        del obj["id"]
        del obj["userlastmodified"]
        del obj["lastmodified"]
        del obj["datecreated"]
        del obj["lastdeployed"]
        for pol in obj["policies"]:
            del pol["id"]
            del pol["lastmodified"]

    return tools.json_compare(
        ret_obj1,
        ret_obj2,
        deleted_keys=[
            "id",
            "userlastmodified",
            "lastmodified",
            "datecreated",
            "lastdeployed",
        ],
    )



import logging
import ibmsecurity.isam.fed.federations
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/federations"
requires_modules = ["federation"]
requires_version = "9.0.1.0"


def get_all(isamAppliance, federation_name, count=None, start=None, check_mode=False, force=False):
    """
    Retrieve a list of partners
    """
    ret_obj = ibmsecurity.isam.fed.federations.search(isamAppliance, name=federation_name, check_mode=check_mode,
                                                      force=force)
    fed_id = ret_obj['data']

    if fed_id == {}:
        logger.info("Federation {0}, not found. Skipping get.".format(federation_name))
        return isamAppliance.create_return_object()
    else:
        return _get_all(isamAppliance, fed_id, tools.create_query_string(count=count, start=start))


def _get_all(isamAppliance, fed_id, query_str=''):
    return isamAppliance.invoke_get("Retrieve a list of partners",
                                    "{0}/{1}/partners{2}".format(uri, fed_id, query_str),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, federation_name, partner_name, check_mode=False, force=False):
    """
    Retrieve a partner
    """
    ret_obj = search(isamAppliance, federation_name, partner_name, check_mode=check_mode, force=force)
    fed_id, partner_id = ret_obj['data']

    if fed_id == {} or partner_id is None:
        logger.info(
            "Federation/Partner {0}/{1} had no match, skipping retrieval.".format(federation_name, partner_name))
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, fed_id, partner_id)


def _get(isamAppliance, federation_id, partner_id):
    """
    Internal function to get data using "id" - used to avoid extra calls
    """
    return isamAppliance.invoke_get("Retrieve a partner",
                                    "{0}/{1}/partners/{2}".format(uri, federation_id, partner_id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def search(isamAppliance, federation_name, partner_name, force=False, check_mode=False):
    """
    Search Federation and Partner ID by name
    """
    ret_obj = ibmsecurity.isam.fed.federations.search(isamAppliance, name=federation_name, check_mode=check_mode,
                                                      force=force)
    fed_id = ret_obj['data']
    partner_id = None
    return_obj = isamAppliance.create_return_object()

    if fed_id != {}:
        logger.info("Federation {0} found!".format(federation_name))
        ret_obj = _get_all(isamAppliance, fed_id)
        for obj in ret_obj['data']:
            if obj['name'] == partner_name:
                logger.info(
                    "Found Federation/Partner {0}/{1} - id: {2}".format(federation_name, partner_name, obj['id']))
                partner_id = obj['id']
    else:
        logger.info('Federation {0} not found!'.format(federation_name))

    return_obj['data'] = fed_id, partner_id

    return return_obj


def import_metadata(isamAppliance, federation_name, filename, partner_name=None, check_mode=False, force=False):
    """
    Import a new partner
    """
    ret_obj = search(isamAppliance, federation_name, partner_name, check_mode, force)
    fed_id, partner_id = ret_obj['data']

    if fed_id == {}:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot import data into unknown federation: {0}".format(federation_name))
    else:
        if partner_id is not None:
            logger.info("Federation / Partner {0}/{1} already exists.".format(federation_name, partner_name))
        if force is True or partner_id is None:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                json_data = {}
                # Override partner name in metadata file - if provided
                if partner_name is not None:
                    json_data['name'] = partner_name
                return isamAppliance.invoke_post_files(
                    "Import a new partner",
                    "{0}/{1}/partners/metadata".format(uri, fed_id),
                    [
                        {
                            'file_formfield': 'metadata',
                            'filename': filename,
                            'mimetype': 'application/octet-stream'
                        }
                    ], json_data,
                    requires_modules=requires_modules,
                    requires_version=requires_version)

    return isamAppliance.create_return_object()


def set(isamAppliance, federation_name, partner_name, enabled, role, configuration, templateName=None, new_name=None,
        check_mode=False, force=False):
    """
    Creating or Modifying a Partner
    """
    ret_obj = search(isamAppliance, federation_name, partner_name)
    fed_id, partner_id = ret_obj['data']

    if fed_id == {}:
        logger.info("Federation {0} not found, skipping add/update.".format(federation_name))
    else:
        if partner_id is None:
            # Force the add - we already know partner does not exist
            return add(isamAppliance, federation_name, partner_name, enabled, role, configuration, templateName,
                       check_mode, True)
        else:
            # Update request
            return update(isamAppliance, federation_name, partner_name, enabled, configuration, role, templateName,
                          new_name, check_mode, force)


def add(isamAppliance, federation_name, partner_name, enabled, role, configuration, templateName=None, check_mode=False,
        force=False):
    """
    Create a new partner
    """
    ret_obj = search(isamAppliance, federation_name, partner_name, check_mode, force)
    fed_id, partner_id = ret_obj['data']
    if fed_id == {}:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot add partner into unknown federation: {0}".format(federation_name))
    else:
        if partner_id is not None:
            logger.info("Federation / Partner {0}/{1} already exists.".format(federation_name, partner_name))

        if force is True or partner_id is None:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                json_data = {
                    "name": partner_name,
                    "enabled": enabled,
                    "role": role,
                    "configuration": configuration
                }
                if templateName is not None:
                    json_data['templateName'] = templateName
                return isamAppliance.invoke_post(
                    "Create a new partner",
                    "{0}/{1}/partners".format(uri, fed_id), json_data,
                    requires_modules=requires_modules,
                    requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, federation_name, partner_name, check_mode=False, force=False):
    """
    Delete a partner
    """
    ret_obj = search(isamAppliance, federation_name, partner_name, check_mode=check_mode, force=force)
    fed_id, partner_id = ret_obj['data']
    if fed_id == {}:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot add partner into unknown federation: {0}".format(federation_name))
    else:
        if partner_id is None:
            logger.info("In federation: {0}, Partner {1} no longer exists.".format(federation_name, partner_name))
        else:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_delete(
                    "Delete a partner",
                    "{0}/{1}/partners/{2}".format(uri, fed_id, partner_id),
                    requires_modules=requires_modules,
                    requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, federation_name, partner_name, enabled, configuration, role=None, templateName=None,
           new_name=None, check_mode=False, force=False):
    """
    Update a specific partner
    """
    fed_id, partner_id, update_required, json_data = _check(isamAppliance, federation_name, partner_name, enabled,
                                                            configuration, role, templateName, new_name)
    if fed_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot update partner in unknown federation: {0}".format(federation_name))

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specific partner",
                "{0}/{1}/partners/{2}".format(uri, fed_id, partner_id), json_data,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, federation_name, partner_name, enabled, configuration, role, templateName, new_name):
    """
    Check and return True if update needed
    """
    update_required = False
    json_data = {
        "enabled": enabled,
        "configuration": configuration
    }
    ret_obj = search(isamAppliance, federation_name, partner_name)
    fed_id, partner_id = ret_obj['data']
    if fed_id == {} or partner_id is None:
        logger.info("Federation/Partner not found, returning no update required.")
        return None, None, update_required, json_data
    else:
        ret_obj = get(isamAppliance, federation_name, partner_name)
        if ret_obj['data'] == {}:
            from ibmsecurity.appliance.ibmappliance import IBMError
            raise IBMError("999", "Search() found partner, but get() failed. Something seriously wrong!")
        else:
            if new_name is not None:
                json_data['name'] = new_name
            else:
                json_data['name'] = partner_name
            # Only added parameters passed, remove everything else
            if role is not None:
                json_data['role'] = role
            else:
                del ret_obj['data']['role']
            if templateName is not None:
                json_data['templateName'] = templateName
            else:
                # May not exist so skip any exceptions when deleting
                try:
                    del ret_obj['data']['templateName']
                except:
                    pass
            # Check to see if configuration data contains mapping rule reference id
            # So special logic to see if mapping rule has changed
            new_map_rule_id, new_map_rule = None, None
            exist_map_rule_id, exist_map_rule = None, None
            try:
                new_map_rule_id = configuration['identityMapping']['properties']['identityMappingRuleReference']
            except:
                pass  # Ignore any lookup errors because of missing key/values
            try:
                new_map_rule = configuration['identityMapping']['properties']['identityMappingRule']
            except:
                pass  # Ignore any lookup errors because of missing key/values
            try:
                exist_map_rule_id = ret_obj['data']['configuration']['identityMapping']['properties'][
                    'identityMappingRuleReference']
            except:
                pass  # Ignore any lookup errors because of missing key/values
            try:
                exist_map_rule = ret_obj['data']['configuration']['identityMapping']['properties'][
                    'identityMappingRule']
            except:
                pass  # Ignore any lookup errors because of missing key/values
            logger.debug("New Mapping data: {0}/{1}".format(new_map_rule_id, new_map_rule))
            logger.debug("Existing Mapping data: {0}/{1}".format(exist_map_rule_id, exist_map_rule))
            if new_map_rule_id is not None:
                # Mapping rule id provided in update, check if we need to extract actual rules and compare
                if exist_map_rule_id is None and exist_map_rule is not None:
                    import ibmsecurity.isam.aac.mapping_rules
                    new_map_obj = ibmsecurity.isam.aac.mapping_rules._get(isamAppliance, new_map_rule_id)
                    if new_map_obj['data'] != {}:
                        if new_map_obj['data']['content'].strip() != exist_map_rule.strip():
                            logger.info("Mapping Rule is different, need to update.")
                            return fed_id, True, json_data
                # Allow mapping rule id to be compared - not actual rules contents
                elif exist_map_rule_id is not None:
                    logger.debug("Comapring mapping rule ids and ignoring actual rule contents.")
                    if exist_map_rule is not None:
                        del ret_obj['data']['configuration']['identityMapping']['properties']['identityMappingRule']
                    if new_map_rule is not None:
                        del json_data['configuration']['identityMapping']['properties']['identityMappingRule']
            del ret_obj['data']['id']
            import ibmsecurity.utilities.tools
            sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
            logger.debug("Sorted input: {0}".format(sorted_json_data))
            sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
            logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
            if sorted_ret_obj != sorted_json_data:
                logger.info("Changes detected, update needed.")
                update_required = True
            json_data['configuration'] = configuration

    return fed_id, partner_id, update_required, json_data



import logging
from ibmsecurity.isam.base.remote_syslog import forwarder
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get_all(isamAppliance, server, port, protocol, check_mode=False, force=False):
    """
    Retrieve the remote syslog forwarding sources
    """
    ret_obj = forwarder.get(isamAppliance=isamAppliance, server=server, port=port, protocol=protocol,
                            check_mode=check_mode, force=force)

    try:
        ret_obj['data'] = ret_obj['data']['sources']
    except:
        ret_obj['data'] = {}

    return ret_obj


def _find_forwarder_source(ret_obj, server, port, protocol, name):
    '''
    Will return None or the forwarder object found
    index returned will be length of array if no match
    '''
    existing_forwarder, i = forwarder._find_forwarder(ret_obj, server, port, protocol)

    existing_forwarder_source, j = None, 0
    if existing_forwarder is not None:
        try:
            for src in existing_forwarder['sources']:
                if src['name'] == name:
                    existing_forwarder_source = src
                    break
                j += 1
        except:
            pass

    return existing_forwarder, i, existing_forwarder_source, j


def get(isamAppliance, server, port, protocol, name, check_mode=False, force=False):
    """
    Retrieve the remote syslog forwarding source
    """
    ret_obj = forwarder.get_all(isamAppliance=isamAppliance, check_mode=check_mode, force=force)

    if isinstance(port, basestring):
        port = int(port)

    existing_forwarder, i, existing_forwarder_source, j = _find_forwarder_source(ret_obj, server, port, protocol, name)

    return_obj = isamAppliance.create_return_object()
    return_obj['data'] = existing_forwarder_source

    return return_obj


def delete(isamAppliance, server, port, protocol, name, check_mode=False, force=False):
    """
    Delete a remote syslog forwarding source
    """
    ret_obj = forwarder.get_all(isamAppliance=isamAppliance, check_mode=check_mode, force=force)
    json_to_post = ret_obj['data']

    if isinstance(port, basestring):
        port = int(port)

    existing_forwarder, i, existing_forwarder_source, j = _find_forwarder_source(ret_obj, server, port, protocol, name)
    if existing_forwarder is not None and existing_forwarder_source is not None:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            del json_to_post[i]['sources'][j]
            return forwarder._update_forwarder_policy(isamAppliance, json_to_post)

    return isamAppliance.create_return_object()


def set(isamAppliance, server, port, protocol, name, tag, facility, severity, check_mode=False, force=False):
    """
    Set a remote syslog forwarding source
    """
    ret_obj = forwarder.get_all(isamAppliance=isamAppliance, check_mode=check_mode, force=force)

    if isinstance(port, basestring):
        port = int(port)

    existing_forwarder, i, existing_forwarder_source, j = _find_forwarder_source(ret_obj, server, port, protocol, name)

    json_data = {
        'name': name,
        'tag': tag,
        'facility': facility,
        'severity': severity
    }

    warnings = []
    update_required = False
    if existing_forwarder is None:
        warnings.append(f"Unable to find forwarder for {server} {port} {protocol}")
    else:
        if existing_forwarder_source is None:
            json_to_post = ret_obj['data']
            json_to_post[i]['sources'].append(json_data)
            update_required = True
        else:
            sorted_json_data = tools.json_sort(json_data)
            logger.debug(f"Sorted input: {sorted_json_data}")
            sorted_ret_obj = tools.json_sort(existing_forwarder_source)
            logger.debug(f"Sorted existing data: {sorted_ret_obj}")
            if sorted_ret_obj != sorted_json_data:
                logger.info("Changes detected, update needed.")
                ret_obj['data'][i]['sources'][j] = json_data
                json_to_post = ret_obj['data']
                update_required = True

    if update_required:
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return forwarder._update_forwarder_policy(isamAppliance, json_to_post, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/isam/rsyslog_forwarder/source_names/{}/instances/{}/logs"
requires_modules = None
requires_version = "9.0.2.1"


def get(isamAppliance, instance_name, source_name='webseal', check_mode=False, force=False):
    """
    Retrieve the log files for a particular remote syslog forwarding source and instance
    """
    return isamAppliance.invoke_get(
        "Retrieve the log files for a particular remote syslog forwarding source and instance",
        uri.format(source_name, instance_name), requires_modules=requires_modules, requires_version=requires_version)



import logging
import os.path
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)
uri = "/isam/authzserver"
requires_modules = None
requires_version = None
requires_model = "Appliance"


def get_all(isamAppliance, id, check_mode=False, force=False):
    """
    Retrieve the log file names of an existing instance

    """
    return isamAppliance.invoke_get("Retrieve the log file names of an existing instance",
                                    "{0}/{1}/logging/v1".format(uri, id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version,
                                    requires_model=requires_model)


def get(isamAppliance, id, file_id, size=None, start=None, options=None, check_mode=False, force=False):
    """
    Retrieve the log file snippet of an existing instance

    """
    return isamAppliance.invoke_get("Retrieve the log file snippet of an existing instance",
                                    "{0}/{1}/logging/{2}/v1{3}".format(uri, id, file_id,
                                                                       tools.create_query_string(size=size, start=start,
                                                                                                 options=options)),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version,
                                    requires_model=requires_model)


def export_file(isamAppliance, id, file_id, filepath, check_mode=False, force=False):
    """
    Export the log file of an existing instance
    """

    if os.path.exists(filepath) is True:
        logger.info("File '{0}' already exists.  Skipping export.".format(filepath))
        warnings = ["File '{0}' already exists.  Skipping export.".format(filepath)]
        return isamAppliance.create_return_object(warnings=warnings)

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_get_file(
            "Export the log file of an existing instance",
            "{0}/{1}/logging/{2}/v1?export".format(uri, id, file_id), filepath, requires_model=requires_model
        )


def delete(isamAppliance, id, file_id, check_mode=False, force=False):
    """
    Clear the log file of an existing instance
    """

    file_exists, warnings = _check(isamAppliance, id, file_id)

    if force is True or file_exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Clear the log file of an existing instance",
                "{0}/{1}/logging/{2}/v1".format(uri, id, file_id), requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, id, file_id):
    """
    Check to see if the file_id exists or not
    """
    ret_obj = get_all(isamAppliance, id)
    file_exists, warnings = False, ret_obj['warnings']

    if not warnings:
        for obj in ret_obj['data']:
            if obj['id'] == file_id:
                logger.info("Found file_id '{0}'".format(file_id))
                file_exists = True

    return file_exists, warnings



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/core/cli"
requires_modules = None
requires_version = "9.0.3.0"


def execute(isamAppliance, command, input=None, check_mode=False, force=False, ignore_error=False):
    """
    Run CLI Command
    """
    post_data = {
        'command': command
    }
    if input is not None:
        post_data['input'] = input

    warnings = ["Idempotency checks are not coded for CLI calls.",
                "All CLI calls will be marked as changed whether a change happens or not."]
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True, warnings=warnings)
    else:
        return isamAppliance.invoke_post("Run CLI Command", uri, post_data, ignore_error=ignore_error,
                                         requires_modules=requires_modules, requires_version=requires_version,
                                         warnings=warnings)



import logging

logger = logging.getLogger(__name__)
requires_model = "Appliance"

requires_model = "Appliance"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the cluster manager log file names
    """
    return isamAppliance.invoke_get("Retrieve the cluster manager log file names",
                                    "/isam/cluster/logging/v1", requires_model=requires_model)


def get(isamAppliance, file_id, size=100, start=None, options=None, check_mode=False, force=False):
    """
    Retrieve a log file snippet
    """
    return isamAppliance.invoke_get("Retrieve a log file snippet",
                                    f"/isam/cluster/logging/{file_id}/v1", requires_model=requires_model)


def _check(isamAppliance, file_id):
    check_obj = {'vale': False, 'warnings': ""}

    ret_obj = get(isamAppliance, file_id)

    file_exists, warnings = False, ret_obj['warnings']

    if warnings ==[] and ret_obj['data']['contents'] != '':
        file_exists = True
    return file_exists, warnings


def delete(isamAppliance, file_id, check_mode=False, force=False):
    """
    Clear a log file
    """
    file_exists, warnings = _check(isamAppliance, file_id)

    if force is True or file_exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Clear a log file",
                f"/isam/cluster/logging/{file_id}/v1", requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def export_file(isamAppliance, file_id, filename, check_mode=False, force=False):
    """
    Export a cluster manager log file
    """
    import os.path

    if force is True or os.path.exists(filename) is False:
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Export a cluster manager log file",
                f"/isam/cluster/logging/{file_id}/v1?export",
                filename, requires_model=requires_model)

    return isamAppliance.create_return_object()



import logging
import ibmsecurity.utilities.tools
from ibmsecurity.isam.web.reverse_proxy import junctions

logger = logging.getLogger(__name__)
requires_modules = ["wga"]
requires_version = "10.0.4.0"


def config(isamAppliance, instance_id, hostname='127.0.0.1', port=443,
           junction="/isvaop", enable_mtls=None, mutual_key_label=None, reuse_certs=False,
           reuse_acls=False, load_certificate=True,
           check_mode=False, force=False):
    """
    OAuth2 IBM Security Verify OIDC Provider configuration

    :param isamAppliance:
    :param instance_id:
    :param junction:
    :param hostname:
    :param port:
    :param enable_mtls # new in 10.0.8
    :param mutual_key_label # new in 10.0.8
    :param reuse_certs:
    :param reuse_acls:
    :param check_mode:
    :param force:
    :param load_certificate # new in ?  Default is true
    :return:
    """

    warnings = [
        f"Idempotency logic will check for existence of {junction} junction. Use force=True to override."]
    if force or _check_config(isamAppliance, instance_id, junction) is False:
        json_data = {
            "junction": junction,
            "hostname": hostname,
            "port": port,
            "reuse_certs": reuse_certs,
            "reuse_acls": reuse_acls,
            "load_certificate": load_certificate
        }
        if enable_mtls is not None:
            if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "10.0.8.0") < 0:
                warnings.append(
                    f"Appliance at version: {isamAppliance.facts['version']}, enable_mtls: {enable_mtls} is not supported. Needs 10.0.8.0 or higher. Ignoring enable_mtls for this call.")
            else:
                json_data["enable_mtls"] = enable_mtls
        if mutual_key_label is not None:
            if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "10.0.8.0") < 0:
                warnings.append(
                    f"Appliance at version: {isamAppliance.facts['version']}, mutual_key_label: {mutual_key_label} is not supported. Needs 10.0.8.0 or higher. Ignoring mutual_key_label for this call.")
            else:
                json_data["mutual_key_label"] = mutual_key_label
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                "OAuth2 IBM Security Verify OIDC Provider configuration for a reverse proxy instance",
                f"/wga/reverseproxy/{instance_id}/oauth2_config",
                json_data,
                warnings=warnings,
                requires_modules=requires_modules, requires_version=requires_version)
    return isamAppliance.create_return_object(warnings=warnings)

def _check_config(isamAppliance, instance_id, junction):
    """
    Check if the junction for oauth already created.  This is overly simplistic.

    :param isamAppliance:
    :param instance_id:
    :param junction:
    :return:
    """
    ret_obj = junctions.get_all(isamAppliance, instance_id)

    for j in ret_obj['data']:
        if j['id'] == junction:
            logger.info(f"Junction {junction} was found - hence oauth2 config must have already executed.")
            return True

    return False



import logging
from io import open

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of mapping rules
    """
    return isamAppliance.invoke_get("Retrieve a list of mapping rules",
                                    "/iam/access/v8/mapping-rules")


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific mapping rule
    """
    ret_obj = search(isamAppliance, name=name)
    id = ret_obj['data']

    if id == {}:
        return isamAppliance.create_return_object()
    else:
        return _get(isamAppliance, id)


def _get(isamAppliance, id):
    """
    Internal function to get data using "id" - used to avoid extra calls

    :param isamAppliance:
    :param id:
    :return:
    """
    return isamAppliance.invoke_get("Retrieve a specific mapping rule",
                                    f"/iam/access/v8/mapping-rules/{id}")


def search(isamAppliance, name, check_mode=False, force=False):
    """
    Search mapping rule by name

    :param isamAppliance:
    :param name:
    :param check_mode:
    :param force:
    :return:
    """
    ret_obj = get_all(isamAppliance, check_mode, force)

    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            return_obj['data'] = obj['id']
            logger.debug(f"Found id: {obj['id']}")
            return_obj['rc'] = 0

    return return_obj


def set(isamAppliance, name, category, filename=None, content=None, upload_filename=None, check_mode=False,
        force=False):
    """
    Creating or Modifying an Mapping Rule
    """

    if _check(isamAppliance, name=name) is False:
        # Force the add - we already know connection does not exist
        return add(isamAppliance, name=name, filename=filename, content=content, category=category, upload_filename=upload_filename,
                   check_mode=check_mode, force=True)
    else:
        # Update request
        return update(isamAppliance, name=name, content=content, upload_filename=upload_filename, check_mode=check_mode, force=force)


def add(isamAppliance, name, filename=None, content=None, category=None, upload_filename=None, check_mode=False, force=False):
    """
    Add a mapping rule
    """

    if force is True or _check(isamAppliance, name) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if content is None or content == '':
                if upload_filename is None or upload_filename == '':
                    return isamAppliance.create_return_object(
                        warnings="Need to pass content or upload_filename for set() to work.")
                else:
                    with open(upload_filename, 'r') as contentFile:
                        content = contentFile.read()
            if filename is None or filename == '':
                if upload_filename is None or upload_filename == '':
                    return isamAppliance.create_return_object(
                        warnings="Need to pass filename or upload_filename for set() to work.")
                else:
                    filename = _extract_filename(upload_filename)
            return isamAppliance.invoke_post(
                "Add a mapping rule",
                "/iam/access/v8/mapping-rules",
                {
                    "name": name,
                    "fileName": filename,
                    "content": content,
                    "category": category
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a mapping rule
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a mapping rule",
                f"/iam/access/v8/mapping-rules/{ret_obj['data']}")

    return isamAppliance.create_return_object()


def update(isamAppliance, name, content=None, upload_filename=None, check_mode=False, force=False):
    """
    Update a specified mapping rule
    """

    if content is None or content == '':
        if upload_filename is None or upload_filename == '':
            return isamAppliance.create_return_object(
                warnings="Need to pass content or upload_filename for set() to work.")
        else:
            with open(upload_filename, 'r') as contentFile:
                content = contentFile.read()

    update_required = False
    ret_obj = search(isamAppliance, name)
    id = ret_obj['data']
    if force is False:
        if id:
            ret_obj_content = _get(isamAppliance, id)
            # Having to strip whitespace to get a good comparison (suspect carriage returns added after save happens)
            if (ret_obj_content['data']['content']).strip() != (content).strip():
                update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified mapping rule",
                f"/iam/access/v8/mapping-rules/{id}",
                {
                    'content': content
                })

    return isamAppliance.create_return_object()


def export_file(isamAppliance, name, filename, check_mode=False, force=False):
    """
    Export a specific mapping rule
    """
    import os.path
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or (ret_obj['data'] != {} and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            id = ret_obj['data']
            return isamAppliance.invoke_get_file(
                "Export a specific mapping rule",
                f"/iam/access/v8/mapping-rules/{id}/file/",
                filename)

    return isamAppliance.create_return_object()


def upload(isamAppliance, name, upload_filename, filename=None, category="OAUTH", check_mode=False, force=False):
    """
    Import a new mapping rule
    """
    if force is True or _check(isamAppliance, name) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if filename is None:
                filename = _extract_filename(upload_filename)

            return isamAppliance.invoke_post_files(
                "Import a new mapping rule",
                "/iam/access/v8/mapping-rules",
                [
                    {
                        'file_formfield': 'file',
                        'filename': upload_filename,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {
                    "name": name,
                    "filename": filename,
                    "category": category
                })

    return isamAppliance.create_return_object()


def _extract_filename(upload_filename):
    """
    Extract filename from fully qualified path to use if no filename provided
    """
    import os.path
    return os.path.basename(upload_filename)


def import_file(isamAppliance, name, filename, check_mode=False, force=False):
    """
    Import a mapping rule (replace)
    """
    update_required = False
    if force is False:
        ret_obj = search(isamAppliance, name)
        if ret_obj['data'] != {}:
            ret_obj_content = _get(isamAppliance, ret_obj['data'])
            with open(filename, 'r') as infile:
                content = infile.read()
            if (ret_obj_content['data']['content']).strip() != content.strip():
                update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Import a mapping rule (replace)",
                f"/iam/access/v8/mapping-rules/{ret_obj['data']}/file",
                [
                    {
                        'file_formfield': 'file',
                        'filename': filename,
                        'mimetype': 'application/file'
                    }
                ],
                {})

    return isamAppliance.create_return_object()


def _check(isamAppliance, name):
    """
    Check if Mapping Rules already exists
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['name'] == name:
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Mapping Rules between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging


class User:
    super_user = None

    def __init__(self, password, username=None):
        self.logger = logging.getLogger(__name__)
        self.logger.debug('Creating a user')

        if username is None:
            self.username = self.super_user
            self.logger.debug('Creating a super-user')
        else:
            self.username = username
        self.password = password



import logging

logger = logging.getLogger(__name__)


def get(isamAppliance, userid, check_mode=False, force=False):
    """
    Get grants by userid
    """
    return isamAppliance.invoke_get("Get grants by userid",
                                    f"/iam/access/v8/grants/userIds/{userid}")


def get_recent(isamAppliance, userid, timestamp, token_type='refresh_token', check_mode=False, force=False):
    """
    Get grants by userid and tokens more recent than given timestamp, also pass back any other tokens found

    token_type will check refresh tokens and can be changed or ignored by passing None

    other tokens could include recent access tokens (not refresh tokens)
    """
    ret_obj = get(isamAppliance=isamAppliance, userid=userid)

    recent_tokens = []
    other_tokens = []
    for attrbs in ret_obj['data']:
        for tok in attrbs['tokens']:
            if tok['dateCreated'] > timestamp and (tok['subType'] == token_type or token_type is None):
                recent_tokens.append(tok)
            else:
                other_tokens.append(tok)

    new_ret_obj = isamAppliance.create_return_object()
    new_ret_obj['data']['recent'] = recent_tokens
    new_ret_obj['data']['other'] = other_tokens

    return new_ret_obj


def delete(isamAppliance, userid, check_mode=False, force=False):
    """
    Delete grants by userid
    """
    if force is True or _check(isamAppliance, userid) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Delete grants by userid",
                                               f"/iam/access/v8/grants/userIds/{userid}")

    return isamAppliance.create_return_object()


def _check(isamAppliance, userid):
    try:
        ret_obj = get(isamAppliance, userid)
        if len(ret_obj['data']) > 0:
            return True
    except:
        pass

    return False



import logging

logger = logging.getLogger(__name__)

requires_model = "Docker"


def publish_changes(isamAppliance, check_mode=False, force=False):
    """
    Publish configuration changes
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put("Publishing configuration changes", "/docker/publish", {},
                                        requires_model=requires_model)


def stop(isamAppliance, check_mode=False, force=False):
    """
    Stopping the configuration container
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put("Stopping the configuration container", "/core/docker/stop", {},
                                        requires_model=requires_model)



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/dynamic_client_migration"
requires_modules = ["mga"]
requires_version = "10.0.0"


def bulk_migration(isamAppliance, definitionName, definitionId, check_mode=False, force=False):
    """
    Dynamic client bulk migration
    :param isamAppliance:
    :param definitionName:
    :param definitionId:
    :param check_mode:
    :param force:
    :return:
    """

    return isamAppliance.invoke_post(
        "Dynamic client bulk migration",
        f"{uri}",
        {
            'definitionName': definitionName,
            'definitionId': definitionId
        },
        requires_modules=requires_modules, requires_version=requires_version)


def client_migration(isamAppliance, definitionName, client_id, check_mode=False, force=False):
    """
    Dynamic client bulk migration
    :param isamAppliance:
    :param definitionName:
    :param definitionId:
    :param check_mode:
    :param force:
    :return:
    """

    return isamAppliance.invoke_post(
        "Dynamic client bulk migration",
        f"{uri}/{client_id}",
        {
            'definitionName': definitionName
        },
        requires_modules=requires_modules, requires_version=requires_version)



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

module_uri = "/isam/felb"
requires_modules = None
requires_version = None
requires_model= "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving FELB configuration in full
    """
    return isamAppliance.invoke_get("Retrieving FELB configuration in full", module_uri,
                                    requires_modules=requires_modules, requires_version=requires_version, requires_model=requires_model)


def get_config(isamAppliance, check_mode=False, force=False):
    """
    Retrieving FELB configuration
    """
    return isamAppliance.invoke_get("Retrieving FELB configuration", f"{module_uri}/configuration",
                                    requires_modules=requires_modules, requires_version=requires_version, requires_model=requires_model)


def set(isamAppliance, enabled, debug, ha, logging, ssl, services, attributes,
        check_mode=False, force=False):
    """
    Replacing FELB configuration in full
    """

    if force is False:
        update_required, json_data, warnings = _check(isamAppliance, enabled, debug, ha, logging, ssl, services, attributes)

    if force is True or update_required:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put("Replacing FELB configuration in full", module_uri, json_data,
                                            requires_modules=requires_modules, requires_version=requires_version, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, enabled, debug, ha, logging, ssl, services, attributes):

    update_required = False
    ret_obj = get(isamAppliance)
    warnings=ret_obj['warnings']

    json_data = {
        "enabled": enabled,
        "debug": debug,
        "ha": ha,
        "logging": logging,
        "ssl": ssl,
        "services": services,
        "attributes": attributes
    }
    sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
    logger.debug(f"Sorted input: {sorted_json_data}")
    sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
    logger.debug(f"Sorted existing data: {sorted_ret_obj}")
    if sorted_ret_obj != sorted_json_data:
        logger.info("Changes detected, update needed.")
        update_required = True

    return update_required, json_data, warnings


def export_file(isamAppliance, filename, check_mode=False, force=False):
    """
    Exporting FELB configuration
    """
    import os.path
    if force is True or os.path.exists(filename) is False:
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file("Exporting FELB configuration", f"{module_uri}?export=true",
                                            filename=filename, requires_modules=requires_modules,
                                            requires_version=requires_version, requires_model=requires_model)

    return isamAppliance.create_return_object()


def import_file(isamAppliance, file, check_mode=False, force=False):
    """
    Importing FELB configuration
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_post_files(description="Importing FELB configuration",
                                               uri=module_uri,
                                               fileinfo=[{
                                                   'file_formfield': 'file',
                                                   'filename': file,
                                                   'mimetype': 'application/octet-stream'
                                               }],
                                               data={},
                                               requires_modules=requires_modules, requires_version=requires_version, requires_model=requires_model)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare FELB configuration between 2 appliances
    """

    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

logger = logging.getLogger(__name__)

uri = "/isam/azn_server"
requires_modules = None
requires_version = None
version = "v1"


def get_all(isamAppliance, admin_id, admin_pwd, domain="Default", check_mode=False, force=False):
    """
    Retrieving the list of authorization servers - Ansible parameter pattern
    """
    return isamAppliance.invoke_post("Retrieving all authorization servers",
                                     "{0}".format(uri),
                                     {
                                         "user": admin_id,
                                         "password": admin_pwd,
                                         "domain": domain
                                     },
                                     requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, isamUser, domain="Default", check_mode=False, force=False):
    """
    Retrieving the list of authorization servers
    """
    return isamAppliance.invoke_post("Retrieving all authorization servers",
                                     "{0}".format(uri),
                                     {
                                         "user": isamUser.username,
                                         "password": isamUser.password,
                                         "domain": domain
                                     },
                                     requires_modules=requires_modules, requires_version=requires_version)


def delete(isamAppliance, id, isamUser, domain="Default", check_mode=False, force=False):
    """
    Removing an authorization server
    """
    if force is True or _check(isamAppliance, id, isamUser, domain) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Remove a authorization server",
                                               "{0}/{1}".format(uri, id),
                                               {

                                                   "user": isamUser.username,
                                                   "password": isamUser.password,
                                                   "domain": domain
                                               },
                                               requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, id, isamUser, domain):
    """
    Check if authorization server is already created or not

    :param isamAppliance:
    :return:
    """
    ret_obj = get(isamAppliance, isamUser, domain)

    logger.debug("Looking for existing authorization servers in: {0}".format(ret_obj['data']))
    if ret_obj['data']:
        for obj in ret_obj['data']:
            if obj['id'] == id:
                logger.debug("Found authorization server: {0}".format(id))
                return True

    return False



import logging
import ibmsecurity.isam.fed.federations

logger = logging.getLogger(__name__)


def config(isamAppliance, instance_id, federation_id=None, federation_name=None, hostname='127.0.0.1', port='443', username=None,
           password=None, reuse_certs=False, reuse_acls=False, check_mode=False, force=False):
    """
    Federation configuration for a reverse proxy instance

    :param isamAppliance:
    :param instance_id:
    :param federation_id:
    :param federation_name:
    :param hostname:
    :param port:
    :param username:
    :param password:
    :param reuse_certs:
    :param reuse_acls:
    :param check_mode:
    :param force:
    :return:
    """
    if username is None:
        logger.info("Required parameter username missing. Skipping config.")
        return isamAppliance.create_return_object(warning=["Required parameter username missing. Skipping config."])

    if password is None:
        logger.info("Required parameter password missing. Skipping config.")
        return isamAppliance.create_return_object(warning=["Required parameter password missing. Skipping config."])

    if federation_name is not None:
        ret_obj = ibmsecurity.isam.fed.federations.search(isamAppliance, name=federation_name, check_mode=check_mode,
                                                      force=force)
        federation_id = ret_obj['data']

        if federation_id == {}:
            logger.info("Federation {0}, not found. Skipping config.".format(federation_name))
            return isamAppliance.create_return_object()

    if federation_id is None:
        logger.info("Required parameter federation_id missing. Skipping config.")
        return isamAppliance.create_return_object()

    if force is True or _check(isamAppliance, instance_id, federation_id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Federation configuration for a reverse proxy instance",
                "/wga/reverseproxy/{0}/fed_config".format(instance_id),
                {
                    "runtime": {
                        "hostname": hostname,
                        "port": port,
                        "username": username,
                        "password": password
                    },
                    "federation_id": federation_id,
                    "reuse_certs": reuse_certs,
                    "reuse_acls": reuse_acls
                })

    return isamAppliance.create_return_object()


def unconfig(isamAppliance, instance_id, federation_id=None, federation_name=None, check_mode=False, force=False):
    """
    Federation unconfiguration for a reverse proxy instance

    :param isamAppliance:
    :param instance_id:
    :param federation_id:
    :param federation_name:
    :param check_mode:
    :param force:
    :return:
    """

    if federation_name is not None:
        ret_obj = ibmsecurity.isam.fed.federations.search(isamAppliance, name=federation_name, check_mode=check_mode,
                                                      force=force)
        federation_id = ret_obj['data']

        if federation_id == {}:
            logger.info("Federation {0}, not found. Skipping config.".format(federation_name))
            return isamAppliance.create_return_object()

    if federation_id is None:
        logger.info("Required parameter federation_id missing. Skipping config.")
        return isamAppliance.create_return_object()

    if force is True or _check(isamAppliance, instance_id, federation_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Federation unconfiguration for a reverse proxy instance",
                "/wga/reverseproxy/{0}/fed_config/{1}".format(instance_id, federation_id))

    return isamAppliance.create_return_object()


def _check(isamappliance, instance_id, federation_id):
    # WebSEAL has a stanza that should contain the configured federations
    from ibmsecurity.isam.web.reverse_proxy.configuration import entry
    ret_obj = entry.get_all(isamappliance, reverseproxy_id=instance_id, stanza_id="isam-fed-autocfg")

    # IF there is any exception - i.e. stanza not found return False
    try:
        if federation_id in ret_obj['data']:
            logger.info("federation_id {0} found in reverse_proxy stanza isam-fed-autocfg.".format(federation_id))
            return True
    except:
        pass

    return False



import logging
import ibmsecurity.isam.base.management_authorization.role

logger = logging.getLogger(__name__)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieving the list of groups for an authorization roles
    """
    return isamAppliance.invoke_get("Retrieving the list of groups for an authorization roles",
                                    f"/authorization/roles/{name}/groups/v1")


def set(isamAppliance, name, group_name, type='embedded_ldap', check_mode=False, force=False):
    """
    Add a group to management authorization role
    """
    new_group = True
    ret_obj = ibmsecurity.isam.base.management_authorization.role.get(isamAppliance, name)

    if (ret_obj['data']['groups'] == None):
        ret_obj['data']['groups'] = []
    else:
        for grp in ret_obj['data']['groups']:
            if grp['name'] == group_name:
                if grp['type'] == type:
                    if force is False:
                        return isamAppliance.create_return_object()
                    new_group = False
                else:  # Replace group with new type
                    ret_obj['data']['groups'].remove(grp)
                break

    if new_group is True:
        ret_obj['data']['groups'].append({'name': group_name, 'type': type})

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put(
            "Add group to management authorization role",
            f"/authorization/roles/{name}/v1", ret_obj['data'])


def delete(isamAppliance, name, group_name, check_mode=False, force=False):
    """
    Delete a group from management authorization role
    """
    group_found = False
    ret_obj = ibmsecurity.isam.base.management_authorization.role.get(isamAppliance, name)

    if (ret_obj['data']['groups'] != None):
        for grp in ret_obj['data']['groups']:
            if grp['name'] == group_name:
                group_found = True
                ret_obj['data']['groups'].remove(grp)
                break

    if group_found is False and force is False:
        return isamAppliance.create_return_object()

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put(
            "Delete group from management authorization role",
            f"/authorization/roles/{name}/v1", ret_obj['data'])



import logging
from ibmsecurity.utilities import tools
from ibmsecurity.isvg import notifications

logger = logging.getLogger(__name__)

uri = "/widgets/server"

def get(isvgAppliance, check_mode=False, force=False):
    """
    Get Cluster Manager server status
    """
    return isvgAppliance.invoke_get("Retrieving Cluster Manager server status", uri)


def execute(isvgAppliance, operation="restart", check_mode=False, force=False):
    """
    Execute an operation (start, stop or restart) on Cluster Manager server
    """
    check_value, warnings = _check(isvgAppliance, operation)

    if force is True or check_value is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isvgAppliance.invoke_post(
                "Executing an operation on Cluster Manager server", uri + "/" + operation + "/clustermanager",
                {
                    'operation': operation
                }, warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def _check(isvgAppliance, operation):
    """
    Check if operation needs to be applied to process
    """
    ret_obj = notifications.get(isvgAppliance)
    warnings = ret_obj['warnings']

    if ret_obj['rc'] == 0 and len(ret_obj['data']) > 0 and 'items' in ret_obj['data']:
        for item in ret_obj['data']['items']:
            # code not tested yet with the right message id
            if 'message' in item and item['message'] == 'identity_was_server_restart_reqd' and operation == "restart":
                logger.info("Cluster Manager server requires a restart")
                return True,warnings
    else:
        logger.info(
            "Executing {0} on Cluster Manager server not required or will not work. Use force if needed.".format(
                operation))
        return False,warnings

    return False,warnings



import ibmsecurity.utilities.tools
import logging

logger = logging.getLogger(__name__)

module_uri = "/isam/felb/errorpages/"
requires_modules = None
requires_version = None
requires_model = "Appliance"


def get(isamAppliance, error_page, check_mode=False, force=False):
    """
    Retrieving an error page
    """
    return isamAppliance.invoke_get("Retrieving an error page", f"{module_uri}{error_page}",
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    requires_model=requires_model)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving List of Error Pages
    """
    return isamAppliance.invoke_get("Retrieving List of Error Pages", module_uri,
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    requires_model=requires_model)


def set(isamAppliance, error_page, contents, check_mode=False, force=False):
    """
    Updating an error page
    """

    ret_obj = get(isamAppliance, error_page)
    warnings = ret_obj['warnings']

    if 'contents' in ret_obj['data']:
        ret_contents = ret_obj['data']['contents']
    else:
        ret_contents = False

    if force is True or ret_contents != contents:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put("Updating an error page", f"{module_uri}{error_page}",
                                            {
                                                "contents": contents
                                            }, requires_version=requires_version, requires_modules=requires_modules,
                                            requires_model=requires_model)
    else:
        return isamAppliance.create_return_object(warnings=warnings)


def import_file(isamAppliance, error_page, filename, check_mode=False, force=False):
    """
    Uploading an error page
    """

    check_value, warnings = _check(isamAppliance, error_page, filename)

    if force is True or check_value is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post_files(description="Uploading an error page",
                                                   uri=f"{module_uri}{error_page}",
                                                   fileinfo=[{
                                                       'file_formfield': 'file',
                                                       'filename': filename,
                                                       'mimetype': 'application/octet-stream'
                                                   }],
                                                   data={},
                                                   requires_modules=requires_modules, requires_version=requires_version,
                                                   requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def export_file(isamAppliance, error_page, filename, check_mode=False, force=False):
    """
    Downloading an error page
    """

    import os.path
    if force is True or os.path.exists(error_page) is False:
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Downloading an error page",
                f"{module_uri}{error_page}?export=true",
                filename=filename, requires_model=requires_model
            )

    return isamAppliance.create_return_object()


def _check(isamAppliance, error_page, filename):
    ret_obj = get(isamAppliance, error_page)
    value = True
    warnings = ret_obj['warnings']
    ret_contents = ""

    if ret_obj['data'] != {}:
        ret_contents = ret_obj['data']['contents']

    with open(filename, 'r') as infile:
        contents = infile.read()

    if len(ret_contents) > 0 and len(contents) > 0:
        if ret_contents.strip() == contents.strip():
            value = False
            return value, warnings
        else:
            value = True
            return value, warnings
    else:
        value = True
        return value, warnings


def compare(isamAppliance1, isamAppliance2):
    """
    Compare error pages between two appliances
    """

    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import ibmsecurity.utilities.tools
import ibmsecurity.isam.base.network.interfaces

logger = logging.getLogger(__name__)
requires_model="Appliance"

try:
    basestring
except NameError:
    basestring = (str, bytes)


def add(isamAppliance, label, vlanId, name='', enabled=False, comment='', overrideSubnetChecking=False, bondedTo='',
        bondingMode=None, check_mode=False, force=False):
    """
    Creating a (VLAN) interface
    """
    check_value, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)

    if force is True or check_value is None:
        if warnings != []:
            if "Docker" in warnings[0]:
                return isamAppliance.create_return_object(warnings=warnings)
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post("Creating a (VLAN) interface", "/net/ifaces/",
                                             {
                                                 'name': name,
                                                 'label': label,
                                                 'comment': comment,
                                                 'enabled': enabled,
                                                 'vlanId': vlanId,
                                                 'bondedTo': bondedTo,
                                                 'bondingMode': bondingMode,
                                                 'ipv4': {
                                                     'dhcp': {
                                                         'enabled': False,
                                                         'allowManagement': False,
                                                         'providesDefaultRoute': False,
                                                         'routeMetric': None
                                                     },
                                                     'overrideSubnetChecking': overrideSubnetChecking,
                                                     'addresses': []
                                                 },
                                                 'ipv6': {
                                                     'dhcp': {
                                                         'enabled': False,
                                                         'allowManagement': False
                                                     },
                                                     'addresses': []
                                                 }
                                             }, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, label, vlanId, check_mode=False, force=False):
    """
    Deleting a (VLAN) interface

    NOTE: vlanId needs to be provided, cannot delete default interface with Null vlanId
    """
    ret_obj = None
    if force is False:
        ret_obj, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)

    if force is True or ret_obj is not None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_delete("Deleting a (VLAN) interface", f"/net/ifaces/{ret_obj['uuid']}", requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, name, comment, label, enabled, vlanId=None, bondedTo=None,
           bondingMode='', check_mode=False, force=False):
    """
    Update existing (VLAN) interface
    """
    update_needed = False
    json_data = {}
    if isinstance(enabled, basestring):
        if enabled.lower() == 'true':
            enabled = True
        else:
            enabled = False
    if force is False:
        ret_obj, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)
        if ret_obj is not None:
            del ret_obj['objType']
            del ret_obj['type']
            json_data = {
                'uuid': ret_obj['uuid'],
                'name': name,
                'label': label,
                'comment': comment,
                'enabled': enabled,
                'vlanId': vlanId,
                'bondedTo': bondedTo,
                'bondingMode': bondingMode,
                'ipv4': {
                    'dhcp': ret_obj['ipv4']['dhcp'],
                    'addresses': ret_obj['ipv4']['addresses'],
                },
                'ipv6': {
                    'dhcp': ret_obj['ipv6']['dhcp'],
                    'addresses': ret_obj['ipv6']['addresses'],
                }
            }
            sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
            logger.debug(f"Sorted input: {sorted_json_data}")
            sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj)
            logger.debug(f"Sorted existing data: {sorted_ret_obj}")
            if sorted_ret_obj != sorted_json_data:
                logger.info("Changes detected, update needed.")
                update_needed = True

        if force is True or update_needed is True:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                return ibmsecurity.isam.base.network.interfaces._update_interface(isamAppliance, json_data)

    return isamAppliance.create_return_object(warnings=warnings)



import logging
import ibmsecurity.utilities.tools
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/isam/rsyslog_forwarder"
requires_modules = None
requires_version = "9.0.2.1"
warnings = []

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the current remote syslog forwarding policy
    """
    return isamAppliance.invoke_get("Retrieve the current remote syslog forwarding policy", uri,
                                    requires_modules=requires_modules, requires_version=requires_version)


def _find_forwarder(ret_obj, server, port, protocol):
    '''
    Will return None or the forwarder object found
    index returned will be length of array if no match
    '''
    existing_forwarder = None
    i = 0
    for obj in ret_obj['data']:
        if obj['server'] == server and obj['port'] == port and obj['protocol'] == protocol:
            existing_forwarder = obj
            logger.debug(f"Found Forwarder: {obj}")
            break
        i += 1

    return existing_forwarder, i


def get(isamAppliance, server=None, port=None, protocol=None, id=None, check_mode=False, force=False):
    """
    Retrieve a specific remote syslog forwarder
    needs server/port/protocol OR id
    id takes precedence (so server/port/protocol is ignored if id is passed)
    """
    if id is None:
        ret_obj = get_all(isamAppliance, check_mode, force)

        if isinstance(port, basestring):
            port = int(port)

        return_obj = isamAppliance.create_return_object()
        return_obj['data'], i = _find_forwarder(ret_obj, server, port, protocol)
        warnings = []
        if return_obj['data'] is None:
            warnings.append(f"No entry found for server {server} port {port} and protocol {protocol}.")
            return_obj['warnings'] = warnings

        return return_obj
    else:
        ret_obj = isamAppliance.invoke_get("Retrieve the current remote syslog forwarding policy based on id", f"{uri}/{id}",
                                        ignore_error=True,
                                        requires_modules=requires_modules, requires_version='11.0.2.0')
        rc = ret_obj.get('rc', -1)
        if rc >= 400 and rc < 600:
            warnings.append(f"No entry found for server {id}.  Return code {rc}")
            return isamAppliance.create_return_object(warnings=warnings)
        return ret_obj


def delete(isamAppliance, server=None, port=None, protocol=None, id=None, check_mode=False, force=False):
    """
    Remove a specific remote syslog forwarder
    """
    if id is None:
        ret_obj = get_all(isamAppliance, check_mode, force)

        if isinstance(port, basestring):
            port = int(port)

        existing_forwarder, i = _find_forwarder(ret_obj, server, port, protocol)

        json_to_post = ret_obj['data']
        if existing_forwarder is not None:
            if check_mode:
                return isamAppliance.create_return_object(changed=True)
            else:
                del json_to_post[i]
                return _update_forwarder_policy(isamAppliance, json_to_post)

        return isamAppliance.create_return_object()
    else:
        ret_obj = isamAppliance.invoke_delete(
            "Delete specific rsyslog forwarding policy",
            f"{uri}/{id}",
            ignore_error=True,
            requires_modules=requires_modules, requires_version='11.0.2.0'
        )
        # the policy may not exist
        rc = ret_obj.get('rc', -1)
        if rc == 400:
            warnings.append(f"{ret_obj.get('data', {}).get('message', 'unknown error')}")
            return isamAppliance.create_return_object(warnings=warnings)
        if rc == 500:
            warnings.append(f"Internal error occurred: {ret_obj.get('data', {}).get('message', 'unknown error')}")
            return isamAppliance.create_return_object(warnings=warnings)
        return ret_obj


def set(isamAppliance, server=None, port=None, protocol='udp', id=None, debug=False, keyfile=None, ca_certificate=None,
        client_certificate=None, permitted_peers=None, sources=[], format=None, check_mode=False, force=False):
    """
    Update or create a specific remote syslog forwarder
    """
    warnings = []

    if port and isinstance(port, basestring):
        port = int(port)

    json_data = {
        'server': server,
        'port': port,
        'protocol': protocol,
        'debug': debug,
    }
    update_required = False
    if id is None:
        ret_obj = get_all(isamAppliance, check_mode, force)

        existing_forwarder, i = _find_forwarder(ret_obj, server, port, protocol)
        if existing_forwarder is not None and sources == [] and existing_forwarder['sources'] != sources:
            sources = existing_forwarder['sources']
            warnings.append("No sources provided, using existing sources to set forwarder.")
        json_data['sources'] = sources
        if keyfile is not None and keyfile != '':
            json_data['keyfile'] = keyfile
        if ca_certificate is not None and ca_certificate != '':
            json_data['ca_certificate'] = ca_certificate
        if client_certificate is not None and client_certificate != '':
            json_data['client_certificate'] = client_certificate
        if permitted_peers is not None and permitted_peers != '':
            json_data['permitted_peers'] = permitted_peers

        # Check of the given server/port/protocol exist - if not then we add it
        if existing_forwarder is None or force:
            if format is not None:
                if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.2.0") < 0:
                    warnings.append(f"Appliance at version: {isamAppliance.facts['version']}, format requires 10.0.2.0")
                else:
                    json_data["format"] = format
            json_to_post = ret_obj['data']
            json_to_post.append(json_data)
            update_required = True
            warnings.append("This is a new forwarder")
        else:
            if format is not None:
                if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.2.0") < 0:
                    warnings.append(f"Appliance at version: {isamAppliance.facts['version']}, format requires 10.0.2.0")
                else:
                    json_data["format"] = format
            else:
                ret_obj['data'][i].pop('format', None)
            existing_forwarder.pop('id', None)
            if tools.json_equals(existing_forwarder, json_data, ignore_keys_not_in_new=True, skipkeys=True, sort_keys=True):
                update_required = False
            else:
                update_required = True

        if update_required:
            if check_mode:
                return isamAppliance.create_return_object(changed=True, warnings=warnings)
            else:
                return _update_forwarder_policy(isamAppliance, json_to_post, warnings=warnings)

        return isamAppliance.create_return_object(warnings=warnings)
    else:
        existing_forwarder = get(isamAppliance, id=id)
        if existing_forwarder is None or not existing_forwarder['data']:
           # new ... not sure this is valid
           return set(isamAppliance, server=server, port=port, protocol=protocol, id=None, debug=debug,
               keyfile=keyfile,
               ca_certificate=ca_certificate,
               client_certificate=client_certificate,
               permitted_peers=permitted_peers, sources=sources, format=format, check_mode=check_mode, force=force)
        else:
            existing_forwarder = existing_forwarder['data']
            if existing_forwarder is not None and sources == [] and existing_forwarder['sources'] != sources:
                sources = existing_forwarder['sources']
                warnings.append("No sources provided, using existing sources to set forwarder.")
            if keyfile is not None and keyfile != '':
                json_data['keyfile'] = keyfile
            if ca_certificate is not None and ca_certificate != '':
                json_data['ca_certificate'] = ca_certificate
            if client_certificate is not None and client_certificate != '':
                json_data['client_certificate'] = client_certificate
            if permitted_peers is not None and permitted_peers != '':
                json_data['permitted_peers'] = permitted_peers
            if format is not None:
                json_data["format"] = format
            json_data["id"] = id
            json_data["sources"] = sources

            if tools.json_equals(existing_forwarder, json_data, ignore_keys_not_in_new=True, skipkeys=True, sort_keys=True):
                update_required = False
            else:
                update_required = True

            if update_required:
                if check_mode:
                    return isamAppliance.create_return_object(changed=True, warnings=warnings)
                else:
                    return isamAppliance.invoke_put(
                            "Update the current remote syslog forwarding policy", f"{uri}/{id}",
                            json_to_post, requires_modules=requires_modules,
                            requires_version='11.0.2.0', warnings=warnings)


def _update_forwarder_policy(isamAppliance, json_to_post, warnings=[]):
    return isamAppliance.invoke_put(
        "Update the current remote syslog forwarding policy", uri,
        json_to_post, requires_modules=requires_modules,
        requires_version=requires_version, warnings=warnings)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare forwarder Policies between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import ibmsecurity.isam.base.network.interfaces

logger = logging.getLogger(__name__)
requires_modules = ["mga", "federation"]
requires_model = "Appliance"

def get(isamAppliance, check_mode=False, force=False):
    """
    Get runtime endpoints or listening interfaces
    """
    return isamAppliance.invoke_get("Retrieving runtime listening interfaces",
                                    "/mga/runtime_tuning/v1",
                                    requires_modules=requires_modules,requires_model=requires_model)


def set(isamAppliance, interface, port, secure, check_mode=False, force=False):

    """
    Set a runtime listening interface
    """
    ret_obj = isamAppliance.create_return_object()
    exists = False
    secure_existing = None
    warnings = []
    if force is False:
        exists, secure_existing, id, warnings = _check(isamAppliance, interface, port)

    # Delete if interface has different secure setting
    if exists is True and secure_existing != secure:
        if check_mode is True:
            ret_obj['changed'] = True
        else:
            delete(isamAppliance, interface, port, check_mode, force=True)
        exists = False

    if force or exists is False:
        if check_mode is True:
            ret_obj['changed'] = True
        else:
            return isamAppliance.invoke_post(
                "Setting a runtime listening interface",
                "/mga/runtime_tuning/endpoints/v1",
                {
                    'interface': interface,
                    'port': int(port),
                    'secure': bool(secure)
                },
                requires_modules=requires_modules,requires_model=requires_model)

    ret_obj['warnings'].extend(warnings)

    return ret_obj


def set_by_address(isamAppliance, address, port, secure, check_mode=False, force=False):
    ret_obj = ibmsecurity.isam.base.network.interfaces.get_all(isamAppliance)
    uuid = None

    if "Docker" in ' '.join(ret_obj['warnings']):
        warnings = ret_obj['warnings']
        return isamAppliance.create_return_object(warnings=warnings)

    for intfc in ret_obj['data']['interfaces']:
        for intfc_adr in intfc['ipv4']['addresses']:
            if intfc_adr['address'] == address:
                uuid = f"{intfc['uuid']}.{intfc_adr['uuid']}"
                break
        for intfc_adr in intfc['ipv6']['addresses']:
            if intfc_adr['address'] == address:
                uuid = f"{intfc['uuid']}.{intfc_adr['uuid']}"
                break

    return set(isamAppliance, uuid, port, secure, check_mode, force)


def _check(isamAppliance, interface, port):
    """
    Check listening interface for the runtime
    """
    warnings = []
    ret_obj = get(isamAppliance)

    if "Docker" in ' '.join(ret_obj['warnings']):
        warnings = ret_obj['warnings']

    exists = False
    secure = False
    id = None
    try:
        for endpoint in ret_obj['data']['endpoints']:
            if endpoint['interface'] == interface and int(endpoint['port']) == int(port):
                logger.info("Runtime listening parameter already exists")
                exists = True
                secure = endpoint['secure']
                id = endpoint['id']
                break
            else:
                logger.debug("Runtime listening parameter does not match")
    except:
        logger.info("Runtime listening parameter does not exist")

    return exists, secure, id, warnings


def delete(isamAppliance, interface, port, check_mode=False, force=False):

    """
    Delete a runtime listening interface
    """
    warnings = []
    exists = False
    id = None
    if force is False:
        exists, secure, id, warnings = _check(isamAppliance, interface, port)
    else:
        id = f'{interface}:{port}'

    if force is True or exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(warnings, changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a runtime listening interface",
                f"/mga/runtime_tuning/endpoints/{id}/v1",
                requires_modules=requires_modules,requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare runtime listening interface
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    try:
        endpoint1 = ret_obj1['data']['endpoints']
        for endpoint in endpoint1:
            if endpoint['id'] != 'local-interface' and endpoint['id'] != 'all-application-interfaces':
                del endpoint['id']
                del endpoint['interface']
    except:
        pass

    try:
        endpoint2 = ret_obj2['data']['endpoints']
        for endpoint in endpoint2:
            if endpoint['id'] != 'local-interface' and endpoint['id'] != 'all-application-interfaces':
                del endpoint['id']
                del endpoint['interface']
    except:
        pass

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id', 'interface'])



import logging
import ibmsecurity.isam.base.management_authorization.role

logger = logging.getLogger(__name__)


def set(isamAppliance, name, feature_name, access=None, check_mode=False, force=False):
    """
    Set feature with access in management authorization role
    """
    new_feature = True
    ret_obj = ibmsecurity.isam.base.management_authorization.role.get(isamAppliance, name)

    if (ret_obj['data']['features'] == None):
        ret_obj['data']['features'] = []
    else:
        for ftr in ret_obj['data']['features']:
            if ftr['name'] == feature_name:
                new_feature = False
                if ftr['access'] != access:
                    ftr['access'] = access
                elif force is False:  # Everything matches, if no force return
                    return isamAppliance.create_return_object()
                break

    if new_feature is True:
        ret_obj['data']['features'].append({'name': feature_name, 'access': access})

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put(
            "Add group to management authorization role",
            f"/authorization/roles/{name}/v1", ret_obj['data'])


def delete(isamAppliance, name, feature_name, check_mode=False, force=False):
    """
    Delete a feature from management authorization role
    """
    feature_found = False
    ret_obj = ibmsecurity.isam.base.management_authorization.role.get(isamAppliance, name)

    if (ret_obj['data']['features'] != None):
        for ftr in ret_obj['data']['features']:
            if ftr['name'] == feature_name:
                feature_found = True
                ret_obj['data']['features'].remove(ftr)
                break

    if feature_found is False and force is False:
        return isamAppliance.create_return_object()

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put(
            "Delete feature from management authorization role",
            f"/authorization/roles/{name}/v1", ret_obj['data'])



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/net"
requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the network configuration
    """
    return isamAppliance.invoke_get("Retrieving the network configuration", uri, requires_model=requires_model)



import logging


logger = logging.getLogger(__name__)

uri = "/mga/scim/configuration"
requires_modules = ["mga", "federation"]
requires_version = "9.0.2"


def get(isamAppliance, ldap_connection, ldap_objectclasses=None, check_mode=False, force=False):
    """
    Retrieving the current list of user associated ldap attributes from the configured
    """
    if ldap_objectclasses is None:
        return isamAppliance.invoke_get(
            "Retrieving the current list of user associated ldap attributes from the configured ",
            "{0}/urn:ietf:params:scim:schemas:core:2.0:User/ldap_attributes?ldap_connection={1}".format(
                uri, ldap_connection),
            requires_modules=requires_modules,
            requires_version=requires_version
            )
    else:
        return isamAppliance.invoke_get(
            "Retrieving the current list of user associated ldap attributes from the configured ",
            "{0}/urn:ietf:params:scim:schemas:core:2.0:User/ldap_attributes?ldap_connection={1}&ldap_objectclasses={2}".format(
                uri, ldap_connection, ldap_objectclasses),
            requires_modules=requires_modules,
            requires_version=requires_version
            )



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/kerberos/test"
requires_modules = ['wga']
requires_version = None


def test(isamAppliance, name, password, check_mode=False, force=False):
    """
    Verify Kerberos Authentication of the Web Server Principal
    """
    # Note: Check mode and force are not relevant for running a test
    ret_obj = isamAppliance.invoke_post(
        "Verify Kerberos Authentication of the Web Server Principal", uri,
        {
            "name": name,
            "password": password
        }, requires_modules=requires_modules, requires_version=requires_version)

    # Test will not cause a change
    if ret_obj['changed'] == True:
        ret_obj['changed'] = False

    return ret_obj



import logging
import zipfile
import shutil
import json
# from codecs import ignore_errors

from ibmsecurity.utilities.tools import get_random_temp_dir
from ibmsecurity.utilities.tools import json_equals

logger = logging.getLogger(__name__)
requires_model = "Appliance"

certificate_database_optional_args = (
    "token_label",
    "serial_number",
    "passcode",
    "hsm_type",
    "ip",
    "port",
    "kneti_hash",
    "esn",
    "use_rfs",
    "rfs",
    "rfs_port",
    "rfs_auth",
    "safenet_pw",
    "secondary_ip",
    "secondary_port",
    "secondary_kneti_hash",
    "secondary_esn",
    "update_zip"
)
certificate_database_new_10_05 = (
    "safenet_keystore_list",
    "safenet_primary_keystore",
    "safenet_user"
)
certificate_database_new_10_08 = (
    "serial_number"
)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get list of all certificate databases
    """
    return isamAppliance.invoke_get("Retrieving all current certificate database names",
                                    "/isam/ssl_certificates")


def get(isamAppliance, cert_dbase_id, check_mode=False, force=False):
    """
    Retrieving the SSL certificate database details
    """

    retObj = isamAppliance.invoke_get("Retrieving the SSL certificate database details",
                                    f"/isam/ssl_certificates/{cert_dbase_id}/details",
                                    requires_model=requires_model,
                                    ignore_error=True)
    if retObj.get('rc', 0) == 404:
        return isamAppliance.create_return_object(rc=404, warnings=[f"{cert_dbase_id} does not exist"])
    else:
        return retObj


def create(isamAppliance, kdb_name, type='kdb',
           check_mode=False, force=False,
           **kwargs,
          ):
    """
    Create a certificate database
               token_label=None,
           serial_number=None,
           passcode=None,
           hsm_type=None,
           ip=None, port=None, kneti_hash=None, esn=None,
           secondary_ip=None, secondary_port=None, secondary_kneti_hash=None, secondary_esn=None,
           use_rfs=None,
           rfs=None, rfs_port=None, rfs_auth=None, safenet_user=None, safenet_pw=None,
           update_zip=None,
           safenet_keystore_list=None, safenet_primary_keystore=None,
    TODO: Update zip does not work atm.
    """
    warnings = []
    if force or not _check(isamAppliance, kdb_name):
        logger.debug("Creating new keystore")
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                    "kdb_name": kdb_name,
                    "type": type,
                }
            for k, v in kwargs.items():
               if k in certificate_database_optional_args:
                    json_data[k] = v
               # new in 10.0.5
               if k in certificate_database_new_10_05:
                   if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.5.0") < 0:
                       warnings.append(f"Appliance at version: {isamAppliance.facts['version']}, {k}: {v} is not supported. Needs 10.0.5.0 or higher. Ignoring {k} for this call.")
                   else:
                       json_data[k] = v
               # new in 10.0.8
               if k in certificate_database_new_10_08:
                   if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.8.0") < 0:
                       warnings.append(f"Appliance at version: {isamAppliance.facts['version']}, {k}: {v} is not supported. Needs 10.0.8.0 or higher. Ignoring {k} for this call.")
                   else:
                       json_data[k] = v
               # Logic
               if type != 'p11' or json_data.get("hsm_type", None) not in ("safenet","safenet-ha"):
                   # warnings.append("Serial number, safenet_* are only valid for safenet hsm, removed from input")
                   json_data.pop("serial_number", None)
                   json_data.pop("safenet_user", None)
                   json_data.pop("safenet_pw", None)
                   json_data.pop("safenet_keystore_list", None)
                   json_data.pop("safenet_primary_keystore", None)
               if type != 'p11' or json_data.get("hsm_type", None) != "ncipher":
                   warnings.append("kneti_hash, port, secondary_*, rfs are only valid for ncipher hsm, removed from input")
                   json_data.pop("port", None)
                   json_data.pop("kneti_hash", None)
                   json_data.pop("secondary_ip", None)
                   json_data.pop("secondary_port", None)
                   json_data.pop("secondary_kneti_hash", None)
                   json_data.pop("secondary_esn", None)
                   json_data.pop("use_rfs", None)
                   json_data.pop("rfs", None)
                   json_data.pop("rfs_port", None)
                   json_data.pop("rfs_auth", None)

            retObj = isamAppliance.invoke_post( f"Creating certificate database {kdb_name}",
                  "/isam/ssl_certificates",
                  json_data,
                  warnings=warnings,
                  ignore_error=True
               )
            if retObj.get("rc", 0) == 400:
                   warnings.append(f"Invalid type (you need to install an extension to support network hsm {type})")
                   return isamAppliance.create_return_object(warnings=warnings)
            elif retObj.get("rc", 0) > 400 and retObj.get("rc", 0) < 500:
                   warnings.append(f"HTTP error !")
                   return isamAppliance.create_return_object(warnings=warnings)
            elif retObj.get("rc", 0) >= 500 and retObj.get("rc", 0) < 600:
                   warnings.append(f"Server error !")
                   return isamAppliance.create_return_object(warnings=warnings)
            else:
                   return retObj

    return isamAppliance.create_return_object()


def delete(isamAppliance, cert_dbase_id, check_mode=False, force=False):
    """
    Delete a certificate database
    """
    if force is True or _check(isamAppliance, cert_dbase_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a certificate database",
                f"/isam/ssl_certificates/{cert_dbase_id}")

    return isamAppliance.create_return_object()


def export_db(isamAppliance, cert_id, filename, check_mode=False, force=False):
    """
    Export a certificate database
    """
    import os.path

    if force is True or (_check(isamAppliance, cert_id) is True and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Export a certificate database",
                f"/isam/ssl_certificates/{cert_id}?export",
                filename)

    return isamAppliance.create_return_object()


def import_db(isamAppliance, kdb, stash, zip=None, check_mode=False, force=False):
    """
    Import certificate database
    """
    # Grab the filename to use as identifier (strip path and extension)
    import os.path

    tmpdir=None
    if zip != None:
        with zipfile.ZipFile(zip,"r") as zip_ref:
            tmpdir = get_random_temp_dir()
            zip_ref.extractall(tmpdir)
            kdb = tmpdir + '/' + kdb
            stash = tmpdir + '/' + stash

    kdb_id = os.path.basename(kdb)
    kdb_id = os.path.splitext(kdb_id)[0]

    try:
        if force is True or _check(isamAppliance, kdb_id) is False:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_post_files(
                    "Import certificate database",
                    "/isam/ssl_certificates",
                    [
                        {
                            'file_formfield': 'kdb',
                            'filename': kdb,
                            'mimetype': 'application/octet-stream'
                        },
                        {
                            'file_formfield': 'stash',
                            'filename': stash,
                            'mimetype': 'application/octet-stream'
                        }
                    ],
                    {})
    finally:
        if tmpdir != None:
            shutil.rmtree(tmpdir)
    return isamAppliance.create_return_object()


def rename(isamAppliance, cert_id, new_name, check_mode=False, force=False):
    """
    Rename a certificate database
    """
    if force is True or _check(isamAppliance, cert_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Renaming a certificate database",
                f"/isam/ssl_certificates/{cert_id}",
                {
                    "new_name": new_name
                })

    return isamAppliance.create_return_object()


def set(isamAppliance, cert_id, description=None, type="kdb", check_mode=False, force=False, **kwargs):
    """
    Set description for a certificate database (type="kdb)"
    Update network certificate database (type="p11")
    """
    warnings = []
    desc_match = True  # This will remain True even when cert db is not found!

    if type in ("p12", "kdb"):
        if not force:
            if description is None:
                desc_match = True
            else:
                ret_obj = get_all(isamAppliance)
                for certdb in ret_obj['data']:
                    if certdb['id'] == cert_id:
                        if certdb['description'] != description:
                            desc_match = False
                        break

        if force or not desc_match:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_put(
                    "Set description for a certificate database",
                    f"/isam/ssl_certificates/{cert_id}",
                    {
                        "description": description
                    })
    else:
        # type = p11
        json_data = {}
        for k, v in kwargs.items():
            if k in certificate_database_optional_args:
                json_data[k] = v
            # new in 10.0.5
            if k in certificate_database_new_10_05:
                if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.5.0") < 0:
                    warnings.append(f"Appliance at version: {isamAppliance.facts['version']}, {k}: {v} is not supported. Needs 10.0.5.0 or higher. Ignoring {k} for this call.")
                else:
                    json_data[k] = v
            # new in 10.0.8
            if k in certificate_database_new_10_08:
                if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts['version'], "10.0.8.0") < 0:
                    warnings.append(f"Appliance at version: {isamAppliance.facts['version']}, {k}: {v} is not supported. Needs 10.0.8.0 or higher. Ignoring {k} for this call.")
                else:
                    json_data[k] = v
        ret_obj = get(isamAppliance, cert_id)
        # Idempotency
        if ret_obj.get("rc", 0) == 404:
            return isamAppliance.create_return_object(warnings=warnings)
        if not json_equals(ret_obj, json_data):
                return isamAppliance.invoke_put(
                    "Updating network certificate database",
                    f"/isam/ssl_certificates/{cert_id}",
                    json_data,
                    warnings=warnings
                )
    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, id):
    """
    Check if certificate database already exists
    """
    ret_obj = get_all(isamAppliance)
    logger.debug(f"\nCHECKING IF {id} EXISTS")
    for certdb in ret_obj['data']:
        logger.info("Cert databases id " + certdb['id'])
        if certdb['id'] == id:
            return True
    logger.debug(f"\n{id} DOES NOT EXIST\n")
    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare certificate databases between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    # Version is last modified time of file - not worth comparing
    for cert_db in ret_obj1['data']:
        del cert_db['version']
    for cert_db in ret_obj2['data']:
        del cert_db['version']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['version'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)
requires_model = "Appliance"

def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve runtime component status
    """
    requires_model = None
    return isamAppliance.invoke_get("Retrieving web runtime component status",
                                    "/isam/runtime_components/",requires_model=requires_model)


def _check(isamAppliance):
    """
    Check if the runtime process is configured or not
    :param isamAppliance:
    :return:
    """
    ret_obj = get(isamAppliance)
    check_value, warnings=False, ret_obj['warnings']

    if warnings == []:
        if ret_obj['data']['modecode'] == '-1':
            check_value = False
            return check_value, warnings
        else:
            check_value = True
            return check_value, warnings
    else:
        return check_value, warnings

def config(isamAppliance, admin_pwd, ps_mode="local", user_registry="local", ldap_host=None, ldap_port=None,
           ldap_dn=None, ldap_pwd=None, ldap_ssl_db=None, ldap_ssl_label=None, ldap_suffix=None, clean_ldap=False,
           domain="Default", admin_cert_lifetime="1460", ssl_compliance="none", isam_host=None, isam_port="7135",
           local_interface_only=None,
           check_mode=False, force=False):
    """
    Configure Runtime Component
    :param isamAppliance:
    :return:
    """
    requires_model = None
    check_value, warnings = _check(isamAppliance)

    if (force is True or check_value is False):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            json_data = {
                          "ps_mode": ps_mode,
                          "user_registry": user_registry,
                          "ldap_host": ldap_host,
                          "ldap_port": ldap_port,
                          "ldap_dn": ldap_dn,
                          "ldap_pwd": ldap_pwd,
                          "ldap_ssl_db": ldap_ssl_db,
                          "ldap_ssl_label": ldap_ssl_label,
                          "ldap_suffix": ldap_suffix,
                          "clean_ldap": clean_ldap,
                          "domain": domain,
                          "admin_pwd": admin_pwd,
                          "admin_cert_lifetime": admin_cert_lifetime,
                          "ssl_compliance": ssl_compliance,
                          "isam_host": isam_host,
                          "isam_port": isam_port
                        }
            if local_interface_only is not None:
                if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "10.0.2") < 0:
                    warnings.append(
                        "Appliance at version: {0}, local_interface_only: {1} is not supported. Needs 10.0.2 or higher. Ignoring local_interface_only for this call.".format(
                            isamAppliance.facts["version"], local_interface_only))
                else:
                    json_data["local_interface_only"] = local_interface_only

            return isamAppliance.invoke_post("Configure web runtime Component", "/isam/runtime_components/",
                                             json_data, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def unconfig(isamAppliance, clean=False, ldap_dn=None, ldap_pwd=None, check_mode=False, force=False):
    """
    Unconfigure existing runtime component
    """
    requires_model = None
    check_value, warnings = _check(isamAppliance)

    if (force is True or check_value is True):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put("Unconfigure web runtime component",
                                            "/isam/runtime_components/RTE",
                                            {
                                                "operation": "unconfigure",
                                                "force": force,
                                                "clean": clean,
                                                "ldap_dn": ldap_dn,
                                                "ldap_pwd": ldap_pwd
                                            },requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def export_configuration(isamAppliance, filename="rte_config.zip",  check_mode=False, force=False):
    """
    Exporting the Runtime Environment configuration
        -filename is file system location to export the file (e.g. /tmp/rte_config.zip)
    """
    import os.path
    if force is True or os.path.exists(os.path.dirname(filename)) is False:
        if check_mode is True: # No point downloading a file if in check_mode
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_get_file(
                description="Exporting the Runtime Environment configuration",
                uri="/isam/runtime_components?action=export",
                filename=filename,
                requires_version="10.0.4")

    return isamAppliance.create_return_object()


def import_config(isamAppliance, migrate_file, check_mode=False, force=False):
    """
    Import or migrate runtime component
    """
    check_value, warnings = _check(isamAppliance)

    if (force is True or check_value is True):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post_files("Import or Migrate web runtime component",
                                                   "/isam/runtime_components/",
                                                   [{
                                                       'file_formfield': 'migrate_file',
                                                       'filename': migrate_file,
                                                       'mimetype': 'application/octet-stream'
                                                   }],
                                                   {},requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def execute(isamAppliance, operation='restart', check_mode=False, force=False):
    """
    Execute an operation on runtime component

    :param isamAppliance:
    :param operation:
    :return:
    """
    check_value, warnings = _check(isamAppliance)

    if (force is False):
        if warnings == []:
            ret_obj = get(isamAppliance)
            if (ret_obj['data']['statuscode'] == '1'):
                logger.info("ISAM web runtime is unconfigured.")
                return isamAppliance.create_return_object()
            if (ret_obj['data']['statuscode'] == '0' and operation == 'start'):
                logger.info("ISAM web runtime is already started.")
                return isamAppliance.create_return_object()
            if (ret_obj['data']['statuscode'] == '2' and operation == 'stop'):
                logger.info("ISAM web runtime is already stopped.")
                return isamAppliance.create_return_object()
        else:
            return isamAppliance.create_return_object(warnings=warnings)

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True,warnings=warnings)
    else:
        return isamAppliance.invoke_put("Execute an operation on web runtime component",
                                        "/isam/runtime_components/",
                                        {
                                            "operation": operation
                                        },requires_model=requires_model)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare web runtime between 2 appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    # Ignore status. since process start/stop status can be different
    del ret_obj1['data']['status']
    del ret_obj1['data']['statuscode']
    del ret_obj2['data']['status']
    del ret_obj2['data']['statuscode']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['status', 'statuscode'])
